import json
import logging
import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.flight as fl
import pyarrow.parquet
from pyarrow import Schema
from pyarrow._flight import FlightCallOptions, FlightDescriptor, Ticket
from pydantic import StrictInt, StrictStr

from feast import OnDemandFeatureView
from feast.arrow_error_handler import arrow_client_error_handling_decorator
from feast.data_source import DataSource
from feast.feature_logging import (
    FeatureServiceLoggingSource,
    LoggingConfig,
    LoggingSource,
)
from feast.feature_view import FeatureView
from feast.infra.offline_stores import offline_utils
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.permissions.auth.auth_type import AuthType
from feast.permissions.auth_model import AuthConfig
from feast.permissions.client.arrow_flight_auth_interceptor import (
    FlightAuthInterceptorFactory,
)
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage

logger = logging.getLogger(__name__)


class FeastFlightClient(fl.FlightClient):
    @arrow_client_error_handling_decorator
    def get_flight_info(
        self, descriptor: FlightDescriptor, options: FlightCallOptions = None
    ):
        return super().get_flight_info(descriptor, options)

    @arrow_client_error_handling_decorator
    def do_get(self, ticket: Ticket, options: FlightCallOptions = None):
        return super().do_get(ticket, options)

    @arrow_client_error_handling_decorator
    def do_put(
        self,
        descriptor: FlightDescriptor,
        schema: Schema,
        options: FlightCallOptions = None,
    ):
        return super().do_put(descriptor, schema, options)

    @arrow_client_error_handling_decorator
    def list_flights(self, criteria: bytes = b"", options: FlightCallOptions = None):
        return super().list_flights(criteria, options)

    @arrow_client_error_handling_decorator
    def list_actions(self, options: FlightCallOptions = None):
        return super().list_actions(options)


def build_arrow_flight_client(
    scheme: str, host: str, port, auth_config: AuthConfig, cert: str = ""
):
    arrow_scheme = "grpc+tcp"
    if scheme == "https":
        logger.info(
            "Scheme is https so going to connect offline server in SSL(TLS) mode."
        )
        arrow_scheme = "grpc+tls"

    kwargs = {}
    if cert:
        with open(cert, "rb") as root_certs:
            kwargs["tls_root_certs"] = root_certs.read()

    if auth_config.type != AuthType.NONE.value:
        middlewares = [FlightAuthInterceptorFactory(auth_config)]
        return FeastFlightClient(
            f"{arrow_scheme}://{host}:{port}", middleware=middlewares, **kwargs
        )

    return FeastFlightClient(f"{arrow_scheme}://{host}:{port}", **kwargs)


class RemoteOfflineStoreConfig(FeastConfigBaseModel):
    type: Literal["remote"] = "remote"

    scheme: Literal["http", "https"] = "http"

    host: StrictStr
    """ str: remote offline store server port, e.g. the host URL for offline store  of arrow flight server. """

    port: Optional[StrictInt] = None
    """ str: remote offline store server port."""

    cert: StrictStr = ""
    """ str: Path to the public certificate when the offline server starts in TLS(SSL) mode. This may be needed if the offline server started with a self-signed certificate, typically this file ends with `*.crt`, `*.cer`, or `*.pem`.
    If type is 'remote', then this configuration is needed to connect to remote offline server in TLS mode. """


class RemoteRetrievalJob(RetrievalJob):
    def __init__(
        self,
        client: FeastFlightClient,
        api: str,
        api_parameters: Dict[str, Any],
        entity_df: Optional[Union[pd.DataFrame, str]] = None,
        table: pa.Table = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        # Initialize the client connection
        self.client = client
        self.api = api
        self.api_parameters = api_parameters
        self.entity_df = entity_df
        self.table = table
        self._metadata = metadata

    # Invoked to realize the Pandas DataFrame
    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        # We use arrow format because it gives better control of the table schema
        return self._to_arrow_internal().to_pandas()

    # Invoked to synchronously execute the underlying query and return the result as an arrow table
    # This is where do_get service is invoked
    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        return _send_retrieve_remote(
            self.api,
            self.api_parameters,
            self.entity_df,
            self.table,
            self.client,
        )

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return []

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    @property
    def full_feature_names(self) -> bool:
        return self.api_parameters["full_feature_names"]

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: bool = False,
        timeout: Optional[int] = None,
    ):
        """
        Arrow flight action is being used to perform the persist action remotely
        """

        api_parameters = {
            "data_source_name": storage.to_data_source().name,
            "allow_overwrite": allow_overwrite,
            "timeout": timeout,
        }

        # Add api parameters to command
        for key, value in self.api_parameters.items():
            api_parameters[key] = value

        api_parameters["retrieve_func"] = self.api

        _call_put(
            api=RemoteRetrievalJob.persist.__name__,
            api_parameters=api_parameters,
            client=self.client,
            table=self.table,
            entity_df=self.entity_df,
        )


class RemoteOfflineStore(OfflineStore):
    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Optional[Union[pd.DataFrame, str]],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RemoteRetrievalJob:
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        client = build_arrow_flight_client(
            scheme=config.offline_store.scheme,
            host=config.offline_store.host,
            port=config.offline_store.port,
            auth_config=config.auth_config,
            cert=config.offline_store.cert,
        )

        feature_view_names = [fv.name for fv in feature_views]
        name_aliases = [fv.projection.name_alias for fv in feature_views]

        api_parameters = {
            "feature_view_names": feature_view_names,
            "feature_refs": feature_refs,
            "project": project,
            "full_feature_names": full_feature_names,
            "name_aliases": name_aliases,
        }

        return RemoteRetrievalJob(
            client=client,
            api=OfflineStore.get_historical_features.__name__,
            api_parameters=api_parameters,
            entity_df=entity_df,
            metadata=_create_retrieval_metadata(feature_refs, entity_df),
        )

    @staticmethod
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        # Initialize the client connection
        client = build_arrow_flight_client(
            scheme=config.offline_store.scheme,
            host=config.offline_store.host,
            port=config.offline_store.port,
            auth_config=config.auth_config,
            cert=config.offline_store.cert,
        )

        api_parameters = {
            "data_source_name": data_source.name,
            "join_key_columns": join_key_columns,
            "feature_name_columns": feature_name_columns,
            "timestamp_field": timestamp_field,
            "created_timestamp_column": created_timestamp_column,
            "start_date": start_date.isoformat() if start_date else None,
            "end_date": end_date.isoformat() if end_date else None,
        }

        return RemoteRetrievalJob(
            client=client,
            api=OfflineStore.pull_all_from_table_or_query.__name__,
            api_parameters=api_parameters,
        )

    @staticmethod
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        # Initialize the client connection
        client = build_arrow_flight_client(
            config.offline_store.scheme,
            config.offline_store.host,
            config.offline_store.port,
            config.auth_config,
            cert=config.offline_store.cert,
        )

        api_parameters = {
            "data_source_name": data_source.name,
            "join_key_columns": join_key_columns,
            "feature_name_columns": feature_name_columns,
            "timestamp_field": timestamp_field,
            "created_timestamp_column": created_timestamp_column,
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
        }

        return RemoteRetrievalJob(
            client=client,
            api=OfflineStore.pull_latest_from_table_or_query.__name__,
            api_parameters=api_parameters,
        )

    @staticmethod
    def write_logged_features(
        config: RepoConfig,
        data: Union[pyarrow.Table, Path],
        source: LoggingSource,
        logging_config: LoggingConfig,
        registry: BaseRegistry,
    ):
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)
        assert isinstance(source, FeatureServiceLoggingSource)

        if isinstance(data, Path):
            data = pyarrow.parquet.read_table(data, use_threads=False, pre_buffer=False)

        # Initialize the client connection
        client = build_arrow_flight_client(
            config.offline_store.scheme,
            config.offline_store.host,
            config.offline_store.port,
            config.auth_config,
            config.offline_store.cert,
        )

        api_parameters = {
            "feature_service_name": source._feature_service.name,
        }

        _call_put(
            api=OfflineStore.write_logged_features.__name__,
            api_parameters=api_parameters,
            client=client,
            table=data,
            entity_df=None,
        )

    @staticmethod
    def offline_write_batch(
        config: RepoConfig,
        feature_view: FeatureView,
        table: pyarrow.Table,
        progress: Optional[Callable[[int], Any]],
    ):
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        # Initialize the client connection
        client = build_arrow_flight_client(
            config.offline_store.scheme,
            config.offline_store.host,
            config.offline_store.port,
            config.auth_config,
            config.offline_store.cert,
        )

        feature_view_names = [feature_view.name]
        name_aliases = [feature_view.projection.name_alias]

        api_parameters = {
            "feature_view_names": feature_view_names,
            "progress": progress,
            "name_aliases": name_aliases,
        }

        _call_put(
            api=OfflineStore.offline_write_batch.__name__,
            api_parameters=api_parameters,
            client=client,
            table=table,
            entity_df=None,
        )

    def validate_data_source(
        self,
        config: RepoConfig,
        data_source: DataSource,
    ):
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        client = build_arrow_flight_client(
            config.offline_store.scheme,
            config.offline_store.host,
            config.offline_store.port,
            config.auth_config,
            config.offline_store.cert,
        )

        api_parameters = {
            "data_source_proto": str(data_source),
        }
        logger.debug(f"validating DataSource {data_source.name}")
        _call_put(
            api=OfflineStore.validate_data_source.__name__,
            api_parameters=api_parameters,
            client=client,
            table=None,
            entity_df=None,
        )

    def get_table_column_names_and_types_from_data_source(
        self, config: RepoConfig, data_source: DataSource
    ) -> Iterable[Tuple[str, str]]:
        assert isinstance(config.offline_store, RemoteOfflineStoreConfig)

        client = build_arrow_flight_client(
            config.offline_store.scheme,
            config.offline_store.host,
            config.offline_store.port,
            config.auth_config,
            config.offline_store.cert,
        )

        api_parameters = {
            "data_source_proto": str(data_source),
        }
        logger.debug(
            f"Calling {OfflineStore.get_table_column_names_and_types_from_data_source.__name__} with {api_parameters}"
        )
        table = _send_retrieve_remote(
            api=OfflineStore.get_table_column_names_and_types_from_data_source.__name__,
            api_parameters=api_parameters,
            client=client,
            table=None,
            entity_df=None,
        )

        logger.debug(
            f"get_table_column_names_and_types_from_data_source for {data_source.name}: {table}"
        )
        return zip(table.column("name").to_pylist(), table.column("type").to_pylist())


def _create_retrieval_metadata(feature_refs: List[str], entity_df: pd.DataFrame):
    entity_schema = _get_entity_schema(
        entity_df=entity_df,
    )

    event_timestamp_col = offline_utils.infer_event_timestamp_from_entity_df(
        entity_schema=entity_schema,
    )

    timestamp_range = _get_entity_df_event_timestamp_range(
        entity_df, event_timestamp_col
    )

    return RetrievalMetadata(
        features=feature_refs,
        keys=list(set(entity_df.columns) - {event_timestamp_col}),
        min_event_timestamp=timestamp_range[0],
        max_event_timestamp=timestamp_range[1],
    )


def _get_entity_schema(entity_df: pd.DataFrame) -> Dict[str, np.dtype]:
    return dict(zip(entity_df.columns, entity_df.dtypes))


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pd.DataFrame, str],
    entity_df_event_timestamp_col: str,
) -> Tuple[datetime, datetime]:
    if not isinstance(entity_df, pd.DataFrame):
        raise ValueError(
            f"Please provide an entity_df of type {type(pd.DataFrame)} instead of type {type(entity_df)}"
        )

    entity_df_event_timestamp = entity_df.loc[
        :, entity_df_event_timestamp_col
    ].infer_objects()
    if pd.api.types.is_string_dtype(entity_df_event_timestamp):
        entity_df_event_timestamp = pd.to_datetime(entity_df_event_timestamp, utc=True)

    return (
        entity_df_event_timestamp.min().to_pydatetime(),
        entity_df_event_timestamp.max().to_pydatetime(),
    )


def _send_retrieve_remote(
    api: str,
    api_parameters: Dict[str, Any],
    entity_df: Optional[Union[pd.DataFrame, str]],
    table: Optional[pa.Table],
    client: FeastFlightClient,
):
    command_descriptor = _call_put(
        api,
        api_parameters,
        client,
        entity_df,
        table,
    )
    return _call_get(client, command_descriptor)


def _call_get(
    client: FeastFlightClient,
    command_descriptor: fl.FlightDescriptor,
):
    flight = client.get_flight_info(command_descriptor)
    ticket = flight.endpoints[0].ticket
    reader = client.do_get(ticket)
    return read_all(reader)


def _call_put(
    api: str,
    api_parameters: Dict[str, Any],
    client: FeastFlightClient,
    entity_df: Optional[Union[pd.DataFrame, str]],
    table: pa.Table,
):
    # Generate unique command identifier
    command_id = str(uuid.uuid4())
    command = {
        "command_id": command_id,
        "api": api,
    }
    # Add api parameters to command
    for key, value in api_parameters.items():
        command[key] = value

    command_descriptor = fl.FlightDescriptor.for_command(
        json.dumps(
            command,
        )
    )

    _put_parameters(command_descriptor, entity_df, table, client)
    return command_descriptor


def _put_parameters(
    command_descriptor: fl.FlightDescriptor,
    entity_df: Optional[Union[pd.DataFrame, str]],
    table: pa.Table,
    client: FeastFlightClient,
):
    updatedTable: pa.Table

    if entity_df is not None:
        updatedTable = pa.Table.from_pandas(entity_df)
    elif table is not None:
        updatedTable = table
    else:
        updatedTable = _create_empty_table()

    writer, _ = client.do_put(command_descriptor, updatedTable.schema)

    write_table(writer, updatedTable)


@arrow_client_error_handling_decorator
def write_table(writer, updated_table: pa.Table):
    writer.write_table(updated_table)
    writer.close()


@arrow_client_error_handling_decorator
def read_all(reader):
    return reader.read_all()


def _create_empty_table():
    schema = pa.schema(
        {
            "key": pa.string(),
        }
    )

    keys = ["mock_key"]

    table = pa.Table.from_pydict(dict(zip(schema.names, keys)), schema=schema)

    return table



import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import Callable, Optional, Union

from tqdm import tqdm

from feast import BatchFeatureView, FeatureView, StreamFeatureView


@dataclass
class MaterializationTask:
    """
    A MaterializationTask represents a unit of data that needs to be materialized from an
    offline store to an online store.
    """

    project: str
    feature_view: Union[BatchFeatureView, StreamFeatureView, FeatureView]
    start_time: datetime
    end_time: datetime
    only_latest: bool = True
    tqdm_builder: Union[None, Callable[[int], tqdm]] = None
    disable_event_timestamp: bool = False


class MaterializationJobStatus(enum.Enum):
    WAITING = 1
    RUNNING = 2
    AVAILABLE = 3
    ERROR = 4
    CANCELLING = 5
    CANCELLED = 6
    SUCCEEDED = 7
    PAUSED = 8
    RETRYING = 9


class MaterializationJob(ABC):
    """
    A MaterializationJob represents an ongoing or executed process that materializes data as per the
    definition of a materialization task.
    """

    task: MaterializationTask

    @abstractmethod
    def status(self) -> MaterializationJobStatus: ...

    @abstractmethod
    def error(self) -> Optional[BaseException]: ...

    @abstractmethod
    def should_be_retried(self) -> bool: ...

    @abstractmethod
    def job_id(self) -> str: ...

    @abstractmethod
    def url(self) -> Optional[str]: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/FeatureService.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from feast.protos.feast.core import FeatureViewProjection_pb2 as feast_dot_core_dot_FeatureViewProjection__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x66\x65\x61st/core/FeatureService.proto\x12\nfeast.core\x1a\x1fgoogle/protobuf/timestamp.proto\x1a&feast/core/FeatureViewProjection.proto\"l\n\x0e\x46\x65\x61tureService\x12,\n\x04spec\x18\x01 \x01(\x0b\x32\x1e.feast.core.FeatureServiceSpec\x12,\n\x04meta\x18\x02 \x01(\x0b\x32\x1e.feast.core.FeatureServiceMeta\"\xa4\x02\n\x12\x46\x65\x61tureServiceSpec\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x33\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32!.feast.core.FeatureViewProjection\x12\x36\n\x04tags\x18\x04 \x03(\x0b\x32(.feast.core.FeatureServiceSpec.TagsEntry\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\r\n\x05owner\x18\x06 \x01(\t\x12\x31\n\x0elogging_config\x18\x07 \x01(\x0b\x32\x19.feast.core.LoggingConfig\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x87\x01\n\x12\x46\x65\x61tureServiceMeta\x12\x35\n\x11\x63reated_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xd1\x08\n\rLoggingConfig\x12\x13\n\x0bsample_rate\x18\x01 \x01(\x02\x12\x45\n\x10\x66ile_destination\x18\x03 \x01(\x0b\x32).feast.core.LoggingConfig.FileDestinationH\x00\x12M\n\x14\x62igquery_destination\x18\x04 \x01(\x0b\x32-.feast.core.LoggingConfig.BigQueryDestinationH\x00\x12M\n\x14redshift_destination\x18\x05 \x01(\x0b\x32-.feast.core.LoggingConfig.RedshiftDestinationH\x00\x12O\n\x15snowflake_destination\x18\x06 \x01(\x0b\x32..feast.core.LoggingConfig.SnowflakeDestinationH\x00\x12I\n\x12\x63ustom_destination\x18\x07 \x01(\x0b\x32+.feast.core.LoggingConfig.CustomDestinationH\x00\x12I\n\x12\x61thena_destination\x18\x08 \x01(\x0b\x32+.feast.core.LoggingConfig.AthenaDestinationH\x00\x12`\n\x1e\x63ouchbase_columnar_destination\x18\t \x01(\x0b\x32\x36.feast.core.LoggingConfig.CouchbaseColumnarDestinationH\x00\x1aS\n\x0f\x46ileDestination\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x1c\n\x14s3_endpoint_override\x18\x02 \x01(\t\x12\x14\n\x0cpartition_by\x18\x03 \x03(\t\x1a(\n\x13\x42igQueryDestination\x12\x11\n\ttable_ref\x18\x01 \x01(\t\x1a)\n\x13RedshiftDestination\x12\x12\n\ntable_name\x18\x01 \x01(\t\x1a\'\n\x11\x41thenaDestination\x12\x12\n\ntable_name\x18\x01 \x01(\t\x1a*\n\x14SnowflakeDestination\x12\x12\n\ntable_name\x18\x01 \x01(\t\x1a\x99\x01\n\x11\x43ustomDestination\x12\x0c\n\x04kind\x18\x01 \x01(\t\x12G\n\x06\x63onfig\x18\x02 \x03(\x0b\x32\x37.feast.core.LoggingConfig.CustomDestination.ConfigEntry\x1a-\n\x0b\x43onfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1aS\n\x1c\x43ouchbaseColumnarDestination\x12\x10\n\x08\x64\x61tabase\x18\x01 \x01(\t\x12\r\n\x05scope\x18\x02 \x01(\t\x12\x12\n\ncollection\x18\x03 \x01(\tB\r\n\x0b\x64\x65stination\"I\n\x12\x46\x65\x61tureServiceList\x12\x33\n\x0f\x66\x65\x61tureservices\x18\x01 \x03(\x0b\x32\x1a.feast.core.FeatureServiceBX\n\x10\x66\x65\x61st.proto.coreB\x13\x46\x65\x61tureServiceProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.FeatureService_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\023FeatureServiceProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_FEATURESERVICESPEC_TAGSENTRY']._options = None
  _globals['_FEATURESERVICESPEC_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION_CONFIGENTRY']._options = None
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION_CONFIGENTRY']._serialized_options = b'8\001'
  _globals['_FEATURESERVICE']._serialized_start=120
  _globals['_FEATURESERVICE']._serialized_end=228
  _globals['_FEATURESERVICESPEC']._serialized_start=231
  _globals['_FEATURESERVICESPEC']._serialized_end=523
  _globals['_FEATURESERVICESPEC_TAGSENTRY']._serialized_start=480
  _globals['_FEATURESERVICESPEC_TAGSENTRY']._serialized_end=523
  _globals['_FEATURESERVICEMETA']._serialized_start=526
  _globals['_FEATURESERVICEMETA']._serialized_end=661
  _globals['_LOGGINGCONFIG']._serialized_start=664
  _globals['_LOGGINGCONFIG']._serialized_end=1769
  _globals['_LOGGINGCONFIG_FILEDESTINATION']._serialized_start=1260
  _globals['_LOGGINGCONFIG_FILEDESTINATION']._serialized_end=1343
  _globals['_LOGGINGCONFIG_BIGQUERYDESTINATION']._serialized_start=1345
  _globals['_LOGGINGCONFIG_BIGQUERYDESTINATION']._serialized_end=1385
  _globals['_LOGGINGCONFIG_REDSHIFTDESTINATION']._serialized_start=1387
  _globals['_LOGGINGCONFIG_REDSHIFTDESTINATION']._serialized_end=1428
  _globals['_LOGGINGCONFIG_ATHENADESTINATION']._serialized_start=1430
  _globals['_LOGGINGCONFIG_ATHENADESTINATION']._serialized_end=1469
  _globals['_LOGGINGCONFIG_SNOWFLAKEDESTINATION']._serialized_start=1471
  _globals['_LOGGINGCONFIG_SNOWFLAKEDESTINATION']._serialized_end=1513
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION']._serialized_start=1516
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION']._serialized_end=1669
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION_CONFIGENTRY']._serialized_start=1624
  _globals['_LOGGINGCONFIG_CUSTOMDESTINATION_CONFIGENTRY']._serialized_end=1669
  _globals['_LOGGINGCONFIG_COUCHBASECOLUMNARDESTINATION']._serialized_start=1671
  _globals['_LOGGINGCONFIG_COUCHBASECOLUMNARDESTINATION']._serialized_end=1754
  _globals['_FEATURESERVICELIST']._serialized_start=1771
  _globals['_FEATURESERVICELIST']._serialized_end=1844
# @@protoc_insertion_point(module_scope)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2018 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Null:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _NullEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Null.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    NULL: _Null.ValueType  # 0

class Null(_Null, metaclass=_NullEnumTypeWrapper): ...

NULL: Null.ValueType  # 0
global___Null = Null

class ValueType(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Enum:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ValueType._Enum.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        INVALID: ValueType._Enum.ValueType  # 0
        BYTES: ValueType._Enum.ValueType  # 1
        STRING: ValueType._Enum.ValueType  # 2
        INT32: ValueType._Enum.ValueType  # 3
        INT64: ValueType._Enum.ValueType  # 4
        DOUBLE: ValueType._Enum.ValueType  # 5
        FLOAT: ValueType._Enum.ValueType  # 6
        BOOL: ValueType._Enum.ValueType  # 7
        UNIX_TIMESTAMP: ValueType._Enum.ValueType  # 8
        BYTES_LIST: ValueType._Enum.ValueType  # 11
        STRING_LIST: ValueType._Enum.ValueType  # 12
        INT32_LIST: ValueType._Enum.ValueType  # 13
        INT64_LIST: ValueType._Enum.ValueType  # 14
        DOUBLE_LIST: ValueType._Enum.ValueType  # 15
        FLOAT_LIST: ValueType._Enum.ValueType  # 16
        BOOL_LIST: ValueType._Enum.ValueType  # 17
        UNIX_TIMESTAMP_LIST: ValueType._Enum.ValueType  # 18
        NULL: ValueType._Enum.ValueType  # 19

    class Enum(_Enum, metaclass=_EnumEnumTypeWrapper): ...
    INVALID: ValueType.Enum.ValueType  # 0
    BYTES: ValueType.Enum.ValueType  # 1
    STRING: ValueType.Enum.ValueType  # 2
    INT32: ValueType.Enum.ValueType  # 3
    INT64: ValueType.Enum.ValueType  # 4
    DOUBLE: ValueType.Enum.ValueType  # 5
    FLOAT: ValueType.Enum.ValueType  # 6
    BOOL: ValueType.Enum.ValueType  # 7
    UNIX_TIMESTAMP: ValueType.Enum.ValueType  # 8
    BYTES_LIST: ValueType.Enum.ValueType  # 11
    STRING_LIST: ValueType.Enum.ValueType  # 12
    INT32_LIST: ValueType.Enum.ValueType  # 13
    INT64_LIST: ValueType.Enum.ValueType  # 14
    DOUBLE_LIST: ValueType.Enum.ValueType  # 15
    FLOAT_LIST: ValueType.Enum.ValueType  # 16
    BOOL_LIST: ValueType.Enum.ValueType  # 17
    UNIX_TIMESTAMP_LIST: ValueType.Enum.ValueType  # 18
    NULL: ValueType.Enum.ValueType  # 19

    def __init__(
        self,
    ) -> None: ...

global___ValueType = ValueType

class Value(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BYTES_VAL_FIELD_NUMBER: builtins.int
    STRING_VAL_FIELD_NUMBER: builtins.int
    INT32_VAL_FIELD_NUMBER: builtins.int
    INT64_VAL_FIELD_NUMBER: builtins.int
    DOUBLE_VAL_FIELD_NUMBER: builtins.int
    FLOAT_VAL_FIELD_NUMBER: builtins.int
    BOOL_VAL_FIELD_NUMBER: builtins.int
    UNIX_TIMESTAMP_VAL_FIELD_NUMBER: builtins.int
    BYTES_LIST_VAL_FIELD_NUMBER: builtins.int
    STRING_LIST_VAL_FIELD_NUMBER: builtins.int
    INT32_LIST_VAL_FIELD_NUMBER: builtins.int
    INT64_LIST_VAL_FIELD_NUMBER: builtins.int
    DOUBLE_LIST_VAL_FIELD_NUMBER: builtins.int
    FLOAT_LIST_VAL_FIELD_NUMBER: builtins.int
    BOOL_LIST_VAL_FIELD_NUMBER: builtins.int
    UNIX_TIMESTAMP_LIST_VAL_FIELD_NUMBER: builtins.int
    NULL_VAL_FIELD_NUMBER: builtins.int
    bytes_val: builtins.bytes
    string_val: builtins.str
    int32_val: builtins.int
    int64_val: builtins.int
    double_val: builtins.float
    float_val: builtins.float
    bool_val: builtins.bool
    unix_timestamp_val: builtins.int
    @property
    def bytes_list_val(self) -> global___BytesList: ...
    @property
    def string_list_val(self) -> global___StringList: ...
    @property
    def int32_list_val(self) -> global___Int32List: ...
    @property
    def int64_list_val(self) -> global___Int64List: ...
    @property
    def double_list_val(self) -> global___DoubleList: ...
    @property
    def float_list_val(self) -> global___FloatList: ...
    @property
    def bool_list_val(self) -> global___BoolList: ...
    @property
    def unix_timestamp_list_val(self) -> global___Int64List: ...
    null_val: global___Null.ValueType
    def __init__(
        self,
        *,
        bytes_val: builtins.bytes = ...,
        string_val: builtins.str = ...,
        int32_val: builtins.int = ...,
        int64_val: builtins.int = ...,
        double_val: builtins.float = ...,
        float_val: builtins.float = ...,
        bool_val: builtins.bool = ...,
        unix_timestamp_val: builtins.int = ...,
        bytes_list_val: global___BytesList | None = ...,
        string_list_val: global___StringList | None = ...,
        int32_list_val: global___Int32List | None = ...,
        int64_list_val: global___Int64List | None = ...,
        double_list_val: global___DoubleList | None = ...,
        float_list_val: global___FloatList | None = ...,
        bool_list_val: global___BoolList | None = ...,
        unix_timestamp_list_val: global___Int64List | None = ...,
        null_val: global___Null.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["bool_list_val", b"bool_list_val", "bool_val", b"bool_val", "bytes_list_val", b"bytes_list_val", "bytes_val", b"bytes_val", "double_list_val", b"double_list_val", "double_val", b"double_val", "float_list_val", b"float_list_val", "float_val", b"float_val", "int32_list_val", b"int32_list_val", "int32_val", b"int32_val", "int64_list_val", b"int64_list_val", "int64_val", b"int64_val", "null_val", b"null_val", "string_list_val", b"string_list_val", "string_val", b"string_val", "unix_timestamp_list_val", b"unix_timestamp_list_val", "unix_timestamp_val", b"unix_timestamp_val", "val", b"val"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["bool_list_val", b"bool_list_val", "bool_val", b"bool_val", "bytes_list_val", b"bytes_list_val", "bytes_val", b"bytes_val", "double_list_val", b"double_list_val", "double_val", b"double_val", "float_list_val", b"float_list_val", "float_val", b"float_val", "int32_list_val", b"int32_list_val", "int32_val", b"int32_val", "int64_list_val", b"int64_list_val", "int64_val", b"int64_val", "null_val", b"null_val", "string_list_val", b"string_list_val", "string_val", b"string_val", "unix_timestamp_list_val", b"unix_timestamp_list_val", "unix_timestamp_val", b"unix_timestamp_val", "val", b"val"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["val", b"val"]) -> typing_extensions.Literal["bytes_val", "string_val", "int32_val", "int64_val", "double_val", "float_val", "bool_val", "unix_timestamp_val", "bytes_list_val", "string_list_val", "int32_list_val", "int64_list_val", "double_list_val", "float_list_val", "bool_list_val", "unix_timestamp_list_val", "null_val"] | None: ...

global___Value = Value

class BytesList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___BytesList = BytesList

class StringList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___StringList = StringList

class Int32List(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___Int32List = Int32List

class Int64List(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___Int64List = Int64List

class DoubleList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___DoubleList = DoubleList

class FloatList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___FloatList = FloatList

class BoolList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[builtins.bool] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___BoolList = BoolList

class RepeatedValue(google.protobuf.message.Message):
    """This is to avoid an issue of being unable to specify `repeated value` in oneofs or maps
    In JSON "val" field can be omitted
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VAL_FIELD_NUMBER: builtins.int
    @property
    def val(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Value]: ...
    def __init__(
        self,
        *,
        val: collections.abc.Iterable[global___Value] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["val", b"val"]) -> None: ...

global___RepeatedValue = RepeatedValue



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

* Copyright 2020 The Feast Authors
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     https://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
"""
import builtins
import collections.abc
import feast.core.DataSource_pb2
import feast.core.Entity_pb2
import feast.core.FeatureService_pb2
import feast.core.FeatureTable_pb2
import feast.core.FeatureView_pb2
import feast.core.InfraObject_pb2
import feast.core.OnDemandFeatureView_pb2
import feast.core.Permission_pb2
import feast.core.Project_pb2
import feast.core.SavedDataset_pb2
import feast.core.StreamFeatureView_pb2
import feast.core.ValidationProfile_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Registry(google.protobuf.message.Message):
    """Next id: 18"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ENTITIES_FIELD_NUMBER: builtins.int
    FEATURE_TABLES_FIELD_NUMBER: builtins.int
    FEATURE_VIEWS_FIELD_NUMBER: builtins.int
    DATA_SOURCES_FIELD_NUMBER: builtins.int
    ON_DEMAND_FEATURE_VIEWS_FIELD_NUMBER: builtins.int
    STREAM_FEATURE_VIEWS_FIELD_NUMBER: builtins.int
    FEATURE_SERVICES_FIELD_NUMBER: builtins.int
    SAVED_DATASETS_FIELD_NUMBER: builtins.int
    VALIDATION_REFERENCES_FIELD_NUMBER: builtins.int
    INFRA_FIELD_NUMBER: builtins.int
    PROJECT_METADATA_FIELD_NUMBER: builtins.int
    REGISTRY_SCHEMA_VERSION_FIELD_NUMBER: builtins.int
    VERSION_ID_FIELD_NUMBER: builtins.int
    LAST_UPDATED_FIELD_NUMBER: builtins.int
    PERMISSIONS_FIELD_NUMBER: builtins.int
    PROJECTS_FIELD_NUMBER: builtins.int
    @property
    def entities(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.Entity_pb2.Entity]: ...
    @property
    def feature_tables(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.FeatureTable_pb2.FeatureTable]: ...
    @property
    def feature_views(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.FeatureView_pb2.FeatureView]: ...
    @property
    def data_sources(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.DataSource_pb2.DataSource]: ...
    @property
    def on_demand_feature_views(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.OnDemandFeatureView_pb2.OnDemandFeatureView]: ...
    @property
    def stream_feature_views(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.StreamFeatureView_pb2.StreamFeatureView]: ...
    @property
    def feature_services(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.FeatureService_pb2.FeatureService]: ...
    @property
    def saved_datasets(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.SavedDataset_pb2.SavedDataset]: ...
    @property
    def validation_references(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.ValidationProfile_pb2.ValidationReference]: ...
    @property
    def infra(self) -> feast.core.InfraObject_pb2.Infra: ...
    @property
    def project_metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ProjectMetadata]:
        """Tracking metadata of Feast by project"""
    registry_schema_version: builtins.str
    """to support migrations; incremented when schema is changed"""
    version_id: builtins.str
    """version id, random string generated on each update of the data; now used only for debugging purposes"""
    @property
    def last_updated(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    @property
    def permissions(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.Permission_pb2.Permission]: ...
    @property
    def projects(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[feast.core.Project_pb2.Project]: ...
    def __init__(
        self,
        *,
        entities: collections.abc.Iterable[feast.core.Entity_pb2.Entity] | None = ...,
        feature_tables: collections.abc.Iterable[feast.core.FeatureTable_pb2.FeatureTable] | None = ...,
        feature_views: collections.abc.Iterable[feast.core.FeatureView_pb2.FeatureView] | None = ...,
        data_sources: collections.abc.Iterable[feast.core.DataSource_pb2.DataSource] | None = ...,
        on_demand_feature_views: collections.abc.Iterable[feast.core.OnDemandFeatureView_pb2.OnDemandFeatureView] | None = ...,
        stream_feature_views: collections.abc.Iterable[feast.core.StreamFeatureView_pb2.StreamFeatureView] | None = ...,
        feature_services: collections.abc.Iterable[feast.core.FeatureService_pb2.FeatureService] | None = ...,
        saved_datasets: collections.abc.Iterable[feast.core.SavedDataset_pb2.SavedDataset] | None = ...,
        validation_references: collections.abc.Iterable[feast.core.ValidationProfile_pb2.ValidationReference] | None = ...,
        infra: feast.core.InfraObject_pb2.Infra | None = ...,
        project_metadata: collections.abc.Iterable[global___ProjectMetadata] | None = ...,
        registry_schema_version: builtins.str = ...,
        version_id: builtins.str = ...,
        last_updated: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        permissions: collections.abc.Iterable[feast.core.Permission_pb2.Permission] | None = ...,
        projects: collections.abc.Iterable[feast.core.Project_pb2.Project] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["infra", b"infra", "last_updated", b"last_updated"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_sources", b"data_sources", "entities", b"entities", "feature_services", b"feature_services", "feature_tables", b"feature_tables", "feature_views", b"feature_views", "infra", b"infra", "last_updated", b"last_updated", "on_demand_feature_views", b"on_demand_feature_views", "permissions", b"permissions", "project_metadata", b"project_metadata", "projects", b"projects", "registry_schema_version", b"registry_schema_version", "saved_datasets", b"saved_datasets", "stream_feature_views", b"stream_feature_views", "validation_references", b"validation_references", "version_id", b"version_id"]) -> None: ...

global___Registry = Registry

class ProjectMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PROJECT_FIELD_NUMBER: builtins.int
    PROJECT_UUID_FIELD_NUMBER: builtins.int
    project: builtins.str
    project_uuid: builtins.str
    def __init__(
        self,
        *,
        project: builtins.str = ...,
        project_uuid: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["project", b"project", "project_uuid", b"project_uuid"]) -> None: ...

global___ProjectMetadata = ProjectMetadata



from abc import ABC
from typing import Optional

from .token_extractor import NoAuthTokenExtractor, TokenExtractor
from .token_parser import NoAuthTokenParser, TokenParser


class AuthManager(ABC):
    """
    The authorization manager offers services to manage authorization tokens from client requests
    to extract user details before injecting them in the security context.
    """

    _token_parser: TokenParser
    _token_extractor: TokenExtractor

    def __init__(self, token_parser: TokenParser, token_extractor: TokenExtractor):
        self._token_parser = token_parser
        self._token_extractor = token_extractor

    @property
    def token_parser(self) -> TokenParser:
        return self._token_parser

    @property
    def token_extractor(self) -> TokenExtractor:
        return self._token_extractor


"""
The possibly empty global instance of `AuthManager`.
"""
_auth_manager: Optional[AuthManager] = None


def get_auth_manager() -> AuthManager:
    """
    Return the global instance of `AuthManager`.

    Raises:
        RuntimeError if the clobal instance is not set.
    """
    global _auth_manager
    if _auth_manager is None:
        raise RuntimeError(
            "AuthManager is not initialized. Call 'set_auth_manager' first."
        )
    return _auth_manager


def set_auth_manager(auth_manager: AuthManager):
    """
    Initialize the global instance of `AuthManager`.
    """

    global _auth_manager
    _auth_manager = auth_manager


class AllowAll(AuthManager):
    """
    An AuthManager not extracting nor parsing the authorization token.
    """

    def __init__(self):
        super().__init__(
            token_extractor=NoAuthTokenExtractor(), token_parser=NoAuthTokenParser()
        )



from datetime import timedelta

import pandas as pd
import pyarrow as pa

from feast.infra.compute_engines.local.backends.base import DataFrameBackend


class PandasBackend(DataFrameBackend):
    def columns(self, df):
        return df.columns.tolist()

    def from_arrow(self, table):
        return table.to_pandas()

    def join(self, left, right, on, how):
        return left.merge(right, on=on, how=how)

    def groupby_agg(self, df, group_keys, agg_ops):
        return (
            df.groupby(group_keys)
            .agg(
                **{
                    alias: pd.NamedAgg(column=col, aggfunc=func)
                    for alias, (func, col) in agg_ops.items()
                }
            )
            .reset_index()
        )

    def filter(self, df, expr):
        return df.query(expr)

    def to_arrow(self, df):
        return pa.Table.from_pandas(df)

    def to_timedelta_value(self, delta: timedelta):
        return pd.to_timedelta(delta)

    def drop_duplicates(self, df, keys, sort_by, ascending: bool = False):
        return df.sort_values(by=sort_by, ascending=ascending).drop_duplicates(
            subset=keys
        )

    def rename_columns(self, df, columns: dict[str, str]):
        return df.rename(columns=columns)



# --------------------------------------------------------------------
# Extends OIDC client auth model with an optional `token` field.
# Works on Pydantic v2-only.
#
# Accepted credential sets (exactly **one** of):
#   1 pre-issued `token`
#   2 `client_secret`            (client-credentials flow)
#   3 `username` + `password` + `client_secret`  (ROPG)
# --------------------------------------------------------------------
from __future__ import annotations

from typing import Literal, Optional

from pydantic import ConfigDict, model_validator

from feast.repo_config import FeastConfigBaseModel


class AuthConfig(FeastConfigBaseModel):
    type: Literal["oidc", "kubernetes", "no_auth"] = "no_auth"


class OidcAuthConfig(AuthConfig):
    auth_discovery_url: str
    client_id: str


class OidcClientAuthConfig(OidcAuthConfig):
    # any **one** of the four fields below is sufficient
    username: Optional[str] = None
    password: Optional[str] = None
    client_secret: Optional[str] = None
    token: Optional[str] = None  # pre-issued `token`

    @model_validator(mode="after")
    def _validate_credentials(cls, values):
        """Enforce exactly one valid credential set."""
        d = values.__dict__ if hasattr(values, "__dict__") else values

        has_user_pass = bool(d.get("username")) and bool(d.get("password"))
        has_secret = bool(d.get("client_secret"))
        has_token = bool(d.get("token"))

        # 1 static token
        if has_token and not (has_user_pass or has_secret):
            return values

        # 2 client_credentials
        if has_secret and not has_user_pass and not has_token:
            return values

        # 3 ROPG
        if has_user_pass and has_secret and not has_token:
            return values

        raise ValueError(
            "Invalid OIDC client auth combination: "
            "provide either\n"
            "  • token\n"
            "  • client_secret (without username/password)\n"
            "  • username + password + client_secret"
        )


class NoAuthConfig(AuthConfig):
    pass


class KubernetesAuthConfig(AuthConfig):
    # Optional user token for users (not service accounts)
    user_token: Optional[str] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2018 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import feast.types.Value_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Field(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    DESCRIPTION_FIELD_NUMBER: builtins.int
    name: builtins.str
    value: feast.types.Value_pb2.ValueType.Enum.ValueType
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """Tags for user defined metadata on a field"""
    description: builtins.str
    """Description of the field."""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        value: feast.types.Value_pb2.ValueType.Enum.ValueType = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        description: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["description", b"description", "name", b"name", "tags", b"tags", "value", b"value"]) -> None: ...

global___Field = Field



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Copyright 2021 The Feast Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import feast.core.DataSource_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import google.protobuf.timestamp_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class SavedDatasetSpec(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    PROJECT_FIELD_NUMBER: builtins.int
    FEATURES_FIELD_NUMBER: builtins.int
    JOIN_KEYS_FIELD_NUMBER: builtins.int
    FULL_FEATURE_NAMES_FIELD_NUMBER: builtins.int
    STORAGE_FIELD_NUMBER: builtins.int
    FEATURE_SERVICE_NAME_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    name: builtins.str
    """Name of the dataset. Must be unique since it's possible to overwrite dataset by name"""
    project: builtins.str
    """Name of Feast project that this Dataset belongs to."""
    @property
    def features(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """list of feature references with format "<view name>:<feature name>" """
    @property
    def join_keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """entity columns + request columns from all feature views used during retrieval"""
    full_feature_names: builtins.bool
    """Whether full feature names are used in stored data"""
    @property
    def storage(self) -> global___SavedDatasetStorage: ...
    feature_service_name: builtins.str
    """Optional and only populated if generated from a feature service fetch"""
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """User defined metadata"""
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        project: builtins.str = ...,
        features: collections.abc.Iterable[builtins.str] | None = ...,
        join_keys: collections.abc.Iterable[builtins.str] | None = ...,
        full_feature_names: builtins.bool = ...,
        storage: global___SavedDatasetStorage | None = ...,
        feature_service_name: builtins.str = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["storage", b"storage"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["feature_service_name", b"feature_service_name", "features", b"features", "full_feature_names", b"full_feature_names", "join_keys", b"join_keys", "name", b"name", "project", b"project", "storage", b"storage", "tags", b"tags"]) -> None: ...

global___SavedDatasetSpec = SavedDatasetSpec

class SavedDatasetStorage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_STORAGE_FIELD_NUMBER: builtins.int
    BIGQUERY_STORAGE_FIELD_NUMBER: builtins.int
    REDSHIFT_STORAGE_FIELD_NUMBER: builtins.int
    SNOWFLAKE_STORAGE_FIELD_NUMBER: builtins.int
    TRINO_STORAGE_FIELD_NUMBER: builtins.int
    SPARK_STORAGE_FIELD_NUMBER: builtins.int
    CUSTOM_STORAGE_FIELD_NUMBER: builtins.int
    ATHENA_STORAGE_FIELD_NUMBER: builtins.int
    @property
    def file_storage(self) -> feast.core.DataSource_pb2.DataSource.FileOptions: ...
    @property
    def bigquery_storage(self) -> feast.core.DataSource_pb2.DataSource.BigQueryOptions: ...
    @property
    def redshift_storage(self) -> feast.core.DataSource_pb2.DataSource.RedshiftOptions: ...
    @property
    def snowflake_storage(self) -> feast.core.DataSource_pb2.DataSource.SnowflakeOptions: ...
    @property
    def trino_storage(self) -> feast.core.DataSource_pb2.DataSource.TrinoOptions: ...
    @property
    def spark_storage(self) -> feast.core.DataSource_pb2.DataSource.SparkOptions: ...
    @property
    def custom_storage(self) -> feast.core.DataSource_pb2.DataSource.CustomSourceOptions: ...
    @property
    def athena_storage(self) -> feast.core.DataSource_pb2.DataSource.AthenaOptions: ...
    def __init__(
        self,
        *,
        file_storage: feast.core.DataSource_pb2.DataSource.FileOptions | None = ...,
        bigquery_storage: feast.core.DataSource_pb2.DataSource.BigQueryOptions | None = ...,
        redshift_storage: feast.core.DataSource_pb2.DataSource.RedshiftOptions | None = ...,
        snowflake_storage: feast.core.DataSource_pb2.DataSource.SnowflakeOptions | None = ...,
        trino_storage: feast.core.DataSource_pb2.DataSource.TrinoOptions | None = ...,
        spark_storage: feast.core.DataSource_pb2.DataSource.SparkOptions | None = ...,
        custom_storage: feast.core.DataSource_pb2.DataSource.CustomSourceOptions | None = ...,
        athena_storage: feast.core.DataSource_pb2.DataSource.AthenaOptions | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["athena_storage", b"athena_storage", "bigquery_storage", b"bigquery_storage", "custom_storage", b"custom_storage", "file_storage", b"file_storage", "kind", b"kind", "redshift_storage", b"redshift_storage", "snowflake_storage", b"snowflake_storage", "spark_storage", b"spark_storage", "trino_storage", b"trino_storage"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["athena_storage", b"athena_storage", "bigquery_storage", b"bigquery_storage", "custom_storage", b"custom_storage", "file_storage", b"file_storage", "kind", b"kind", "redshift_storage", b"redshift_storage", "snowflake_storage", b"snowflake_storage", "spark_storage", b"spark_storage", "trino_storage", b"trino_storage"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["kind", b"kind"]) -> typing_extensions.Literal["file_storage", "bigquery_storage", "redshift_storage", "snowflake_storage", "trino_storage", "spark_storage", "custom_storage", "athena_storage"] | None: ...

global___SavedDatasetStorage = SavedDatasetStorage

class SavedDatasetMeta(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    LAST_UPDATED_TIMESTAMP_FIELD_NUMBER: builtins.int
    MIN_EVENT_TIMESTAMP_FIELD_NUMBER: builtins.int
    MAX_EVENT_TIMESTAMP_FIELD_NUMBER: builtins.int
    @property
    def created_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time when this saved dataset is created"""
    @property
    def last_updated_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Time when this saved dataset is last updated"""
    @property
    def min_event_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Min timestamp in the dataset (needed for retrieval)"""
    @property
    def max_event_timestamp(self) -> google.protobuf.timestamp_pb2.Timestamp:
        """Max timestamp in the dataset (needed for retrieval)"""
    def __init__(
        self,
        *,
        created_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        last_updated_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        min_event_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
        max_event_timestamp: google.protobuf.timestamp_pb2.Timestamp | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp", "max_event_timestamp", b"max_event_timestamp", "min_event_timestamp", b"min_event_timestamp"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["created_timestamp", b"created_timestamp", "last_updated_timestamp", b"last_updated_timestamp", "max_event_timestamp", b"max_event_timestamp", "min_event_timestamp", b"min_event_timestamp"]) -> None: ...

global___SavedDatasetMeta = SavedDatasetMeta

class SavedDataset(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SPEC_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    @property
    def spec(self) -> global___SavedDatasetSpec: ...
    @property
    def meta(self) -> global___SavedDatasetMeta: ...
    def __init__(
        self,
        *,
        spec: global___SavedDatasetSpec | None = ...,
        meta: global___SavedDatasetMeta | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["meta", b"meta", "spec", b"spec"]) -> None: ...

global___SavedDataset = SavedDataset



import logging
import os

import jwt
from kubernetes import client, config
from starlette.authentication import (
    AuthenticationError,
)

from feast.permissions.auth.token_parser import TokenParser
from feast.permissions.user import User

logger = logging.getLogger(__name__)
_namespace_file_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"


class KubernetesTokenParser(TokenParser):
    """
    A `TokenParser` implementation to use Kubernetes RBAC resources to retrieve the user details.
    The assumption is that the request header includes an authorization bearer with the token of the
    client `ServiceAccount`.
    By inspecting the role bindings, this `TokenParser` extracts the associated `Role`s.

    The client `ServiceAccount` is instead used as the user name, together with the current namespace.
    """

    def __init__(self):
        config.load_incluster_config()
        self.v1 = client.CoreV1Api()
        self.rbac_v1 = client.RbacAuthorizationV1Api()
        self.auth_v1 = client.AuthenticationV1Api()

    async def user_details_from_access_token(self, access_token: str) -> User:
        """
        Extract user details from the token using Token Access Review.
        Handles both service account tokens (JWTs) and user tokens (opaque tokens).

        Returns:
            User: Current user, with associated roles, groups, and namespaces.

        Raises:
            AuthenticationError if any error happens.
        """
        # First, try to extract user information using Token Access Review
        groups, namespaces = self._extract_groups_and_namespaces_from_token(
            access_token
        )

        # Try to determine if this is a service account or regular user
        try:
            # Attempt to decode as JWT (for service accounts)
            sa_namespace, sa_name = _decode_token(access_token)
            current_user = f"{sa_namespace}:{sa_name}"
            logger.info(
                f"Request received from ServiceAccount: {sa_name} in namespace: {sa_namespace}"
            )

            intra_communication_base64 = os.getenv("INTRA_COMMUNICATION_BASE64")
            if sa_name is not None and sa_name == intra_communication_base64:
                return User(username=sa_name, roles=[], groups=[], namespaces=[])
            else:
                current_namespace = self._read_namespace_from_file()
                logger.info(
                    f"Looking for ServiceAccount roles of {sa_namespace}:{sa_name} in {current_namespace}"
                )

                # Get roles using existing method
                roles = self.get_roles(
                    current_namespace=current_namespace,
                    service_account_namespace=sa_namespace,
                    service_account_name=sa_name,
                )
                logger.info(f"Roles: {roles}")

                return User(
                    username=current_user,
                    roles=roles,
                    groups=groups,
                    namespaces=namespaces,
                )

        except AuthenticationError as e:
            # If JWT decoding fails, this is likely a user token
            # Use Token Access Review to get user information
            logger.info(f"Token is not a JWT (likely a user token): {e}")

            # Get username from Token Access Review
            username = self._get_username_from_token_review(access_token)
            if not username:
                raise AuthenticationError("Could not extract username from token")

            logger.info(f"Request received from User: {username}")

            # Extract roles for the user from RoleBindings and ClusterRoleBindings
            logger.info(f"Extracting roles for user {username} with groups: {groups}")
            roles = self.get_user_roles(username, groups)

            return User(
                username=username, roles=roles, groups=groups, namespaces=namespaces
            )

    def _read_namespace_from_file(self):
        try:
            with open(_namespace_file_path, "r") as file:
                namespace = file.read().strip()
            return namespace
        except Exception as e:
            raise e

    def get_roles(
        self,
        current_namespace: str,
        service_account_namespace: str,
        service_account_name: str,
    ) -> list[str]:
        """
        Fetches the Kubernetes `Role`s associated to the given `ServiceAccount` in `current_namespace` namespace.

        The running deployment must be granted enough permissions to query for such instances in this namespace.

        Returns:
            list[str]: Name of the `Role`s associated to the service account. No string manipulation is performed on the role name.
        """
        role_bindings = self.rbac_v1.list_namespaced_role_binding(current_namespace)
        roles: set[str] = set()

        for binding in role_bindings.items:
            if binding.subjects is not None:
                for subject in binding.subjects:
                    if (
                        subject.kind == "ServiceAccount"
                        and subject.name == service_account_name
                        and subject.namespace == service_account_namespace
                    ):
                        roles.add(binding.role_ref.name)

        return list(roles)

    def get_user_roles(self, username: str, groups: list[str]) -> list[str]:
        """
        Fetches the Kubernetes `Role`s and `ClusterRole`s associated to the given user and their groups.

        This method checks both namespaced RoleBindings and ClusterRoleBindings for:
        - Direct user assignments
        - Group-based assignments where the user is a member

        Returns:
            list[str]: Names of the `Role`s and `ClusterRole`s associated to the user.
        """
        roles: set[str] = set()

        try:
            # Get all namespaced RoleBindings across all namespaces
            all_role_bindings = self.rbac_v1.list_role_binding_for_all_namespaces()

            for binding in all_role_bindings.items:
                if binding.subjects is not None:
                    for subject in binding.subjects:
                        # Check for direct user assignment
                        if subject.kind == "User" and subject.name == username:
                            roles.add(binding.role_ref.name)
                        # Check for group-based assignment
                        elif subject.kind == "Group" and subject.name in groups:
                            roles.add(binding.role_ref.name)

            # Get all ClusterRoleBindings
            cluster_role_bindings = self.rbac_v1.list_cluster_role_binding()

            for binding in cluster_role_bindings.items:
                if binding.subjects is not None:
                    for subject in binding.subjects:
                        # Check for direct user assignment
                        if subject.kind == "User" and subject.name == username:
                            roles.add(binding.role_ref.name)
                        # Check for group-based assignment
                        elif subject.kind == "Group" and subject.name in groups:
                            roles.add(binding.role_ref.name)

            logger.info(f"Found {len(roles)} roles for user {username}: {list(roles)}")

        except Exception as e:
            logger.error(f"Failed to extract user roles for {username}: {e}")

        return list(roles)

    def _extract_groups_and_namespaces_from_token(
        self, access_token: str
    ) -> tuple[list[str], list[str]]:
        """
        Extract groups and namespaces from the token using Kubernetes Token Access Review.

        Args:
            access_token: The JWT token to analyze

        Returns:
            tuple[list[str], list[str]]: A tuple containing (groups, namespaces)
        """
        try:
            # Create TokenReview object
            token_review = client.V1TokenReview(
                spec=client.V1TokenReviewSpec(token=access_token)
            )
            groups: list[str] = []
            namespaces: list[str] = []

            # Call Token Access Review API
            response = self.auth_v1.create_token_review(token_review)

            if response.status.authenticated:
                # Extract groups and namespaces from the response
                # Groups are in response.status.user.groups, not response.status.groups
                if response.status.user and hasattr(response.status.user, "groups"):
                    groups = response.status.user.groups or []
                else:
                    groups = []

                # Extract namespaces from the user info
                if response.status.user:
                    username = getattr(response.status.user, "username", "") or ""

                    if ":" in username and username.startswith(
                        "system:serviceaccount:"
                    ):
                        # Service account logic - extract namespace from username
                        parts = username.split(":")
                        if len(parts) >= 4:
                            service_account_namespace = parts[
                                2
                            ]  # namespace is the 3rd part
                            namespaces.append(service_account_namespace)

                            # For service accounts, also extract groups that have access to this namespace
                            namespace_groups = self._extract_namespace_access_groups(
                                service_account_namespace
                            )
                            groups.extend(namespace_groups)
                    else:
                        # Regular user logic - extract namespaces from dashboard-permissions RoleBindings
                        user_namespaces = self._extract_user_project_namespaces(
                            username
                        )
                        namespaces.extend(user_namespaces)
                        logger.info(
                            f"Found {len(user_namespaces)} data science projects for user {username}: {user_namespaces}"
                        )

                    # Also check if there are namespace-specific groups
                    for group in groups:
                        if group.startswith("system:serviceaccounts:"):
                            # Extract namespace from service account group
                            parts = group.split(":")
                            if len(parts) >= 3:
                                namespaces.append(parts[2])

                logger.debug(
                    f"Token Access Review successful. Groups: {groups}, Namespaces: {namespaces}"
                )
            else:
                logger.warning(f"Token Access Review failed: {response.status.error}")

        except Exception as e:
            logger.error(f"Failed to perform Token Access Review: {e}")
            # We dont need to extract groups and namespaces from jwt decoding, not ideal for kubernetes auth

        # Remove duplicates
        groups = sorted(list(set(groups)))
        namespaces = sorted(list(set(namespaces)))
        return groups, namespaces

    def _extract_namespace_access_groups(self, namespace: str) -> list[str]:
        """
        Extract groups that have access to a specific namespace by querying RoleBindings and ClusterRoleBindings.

        Args:
            namespace: The namespace to check for group access

        Returns:
            list[str]: List of groups that have access to the namespace
        """
        groups = []
        try:
            # Get RoleBindings in the namespace
            role_bindings = self.rbac_v1.list_namespaced_role_binding(
                namespace=namespace
            )
            for rb in role_bindings.items:
                for subject in rb.subjects or []:
                    if subject.kind == "Group":
                        groups.append(subject.name)
                        logger.debug(
                            f"Found group {subject.name} in RoleBinding {rb.metadata.name}"
                        )

            # Get ClusterRoleBindings that might grant access to this namespace
            cluster_role_bindings = self.rbac_v1.list_cluster_role_binding()
            for crb in cluster_role_bindings.items:
                # Check if this ClusterRoleBinding grants access to the namespace
                if self._cluster_role_binding_grants_namespace_access(crb, namespace):
                    for subject in crb.subjects or []:
                        if subject.kind == "Group":
                            groups.append(subject.name)
                            logger.debug(
                                f"Found group {subject.name} in ClusterRoleBinding {crb.metadata.name}"
                            )

            # Remove duplicates and sort
            groups = sorted(list(set(groups)))
            logger.info(
                f"Found {len(groups)} groups with access to namespace {namespace}: {groups}"
            )

        except Exception as e:
            logger.error(
                f"Failed to extract namespace access groups for {namespace}: {e}"
            )

        return groups

    def _extract_user_project_namespaces(self, username: str) -> list[str]:
        """
        Extract data science project namespaces where a user has been added via dashboard-permissions RoleBindings.

        This method queries all RoleBindings where the user is a subject and filters for
        'dashboard-permissions-*' RoleBindings or 'admin' RoleBindings, which indicate the user has been added to that data science project.

        Args:
            username: The username to search for in RoleBindings

        Returns:
            list[str]: List of namespace names where the user has dashboard or admin permissions
        """
        user_namespaces = []
        try:
            # Query all RoleBindings where the user is a subject
            # This is much more efficient than scanning all namespaces
            all_role_bindings = self.rbac_v1.list_role_binding_for_all_namespaces()

            for rb in all_role_bindings.items:
                # Check if this is a dashboard-permissions RoleBinding
                is_dashboard_permissions = (
                    rb.metadata.name.startswith("dashboard-permissions-")
                    and rb.metadata.labels
                    and rb.metadata.labels.get("opendatahub.io/dashboard") == "true"
                )

                # Check if this is an admin RoleBinding
                is_admin_rolebinding = (
                    rb.role_ref
                    and rb.role_ref.kind == "ClusterRole"
                    and rb.role_ref.name == "admin"
                )

                if is_dashboard_permissions or is_admin_rolebinding:
                    # Check if the user is a subject in this RoleBinding
                    for subject in rb.subjects or []:
                        if subject.kind == "User" and subject.name == username:
                            namespace_name = rb.metadata.namespace
                            user_namespaces.append(namespace_name)
                            rolebinding_type = (
                                "dashboard-permissions"
                                if is_dashboard_permissions
                                else "admin"
                            )
                            logger.debug(
                                f"Found user {username} in {rolebinding_type} RoleBinding "
                                f"{rb.metadata.name} in namespace {namespace_name}"
                            )
                            break  # Found the user in this RoleBinding, no need to check other subjects

            # Remove duplicates and sort
            user_namespaces = sorted(list(set(user_namespaces)))
            logger.info(
                f"User {username} has dashboard or admin permissions in {len(user_namespaces)} namespaces: {user_namespaces}"
            )

        except Exception as e:
            logger.error(
                f"Failed to extract user data science projects for {username}: {e}"
            )

        return user_namespaces

    def _get_username_from_token_review(self, access_token: str) -> str:
        """
        Extract username from Token Access Review.

        Args:
            access_token: The access token to review

        Returns:
            str: The username from the token review, or empty string if not found
        """
        try:
            token_review = client.V1TokenReview(
                spec=client.V1TokenReviewSpec(token=access_token)
            )

            response = self.auth_v1.create_token_review(token_review)

            if response.status.authenticated and response.status.user:
                username = getattr(response.status.user, "username", "") or ""
                logger.debug(f"Extracted username from Token Access Review: {username}")
                return username
            else:
                logger.warning(f"Token Access Review failed: {response.status.error}")
                return ""

        except Exception as e:
            logger.error(f"Failed to get username from Token Access Review: {e}")
            return ""

    def _cluster_role_binding_grants_namespace_access(
        self, cluster_role_binding, namespace: str
    ) -> bool:
        """
        Check if a ClusterRoleBinding grants access to a specific namespace.
        This is a simplified check - in practice, you might need more sophisticated logic.

        Args:
            cluster_role_binding: The ClusterRoleBinding to check
            namespace: The namespace to check access for

        Returns:
            bool: True if the ClusterRoleBinding likely grants access to the namespace
        """
        try:
            # Get the ClusterRole referenced by this binding
            cluster_role_name = cluster_role_binding.role_ref.name
            cluster_role = self.rbac_v1.read_cluster_role(name=cluster_role_name)

            # Check if the ClusterRole has rules that could grant access to the namespace
            for rule in cluster_role.rules or []:
                # Check if the rule applies to namespaces or has wildcard access
                if (
                    rule.resources
                    and ("namespaces" in rule.resources or "*" in rule.resources)
                    and rule.verbs
                    and (
                        "get" in rule.verbs or "list" in rule.verbs or "*" in rule.verbs
                    )
                ):
                    return True

                # Check if the rule has resourceNames that include our namespace
                if rule.resource_names and namespace in rule.resource_names:
                    return True

        except Exception as e:
            logger.debug(
                f"Error checking ClusterRoleBinding {cluster_role_binding.metadata.name}: {e}"
            )

        return False


def _decode_token(access_token: str) -> tuple[str, str]:
    """
    The `sub` portion of the decoded token includes the service account name in the format: `system:serviceaccount:NAMESPACE:SA_NAME`

    Returns:
        str: the namespace name.
        str: the `ServiceAccount` name.
    """
    try:
        decoded_token = jwt.decode(access_token, options={"verify_signature": False})
        if "sub" in decoded_token:
            subject: str = decoded_token["sub"]
            if len(subject.split(":")) != 4:
                raise AuthenticationError(
                    f"Expecting 4 elements separated by : in th subject section, instead of {len(subject.split(':'))}."
                )
            _, _, sa_namespace, sa_name = subject.split(":")
            return (sa_namespace, sa_name)
        else:
            raise AuthenticationError("Missing sub section in received token.")
    except jwt.DecodeError as e:
        raise AuthenticationError(f"Error decoding JWT token: {e}")



import contextlib
import warnings
from dataclasses import asdict
from datetime import datetime, timedelta
from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterator,
    KeysView,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    cast,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from couchbase_columnar.cluster import Cluster
from couchbase_columnar.common.result import BlockingQueryResult
from couchbase_columnar.credential import Credential
from couchbase_columnar.options import ClusterOptions, QueryOptions, TimeoutOptions
from jinja2 import BaseLoader, Environment
from pydantic import StrictFloat, StrictStr

from feast.data_source import DataSource
from feast.errors import InvalidEntityType, ZeroRowsQueryResult
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.registry.base_registry import BaseRegistry
from feast.infra.utils.couchbase.couchbase_utils import normalize_timestamp
from feast.on_demand_feature_view import OnDemandFeatureView
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage

from ... import offline_utils
from ...offline_utils import get_timestamp_filter_sql
from .couchbase_source import (
    CouchbaseColumnarSource,
    SavedDatasetCouchbaseColumnarStorage,
)

# Only prints out runtime warnings once.
warnings.simplefilter("once", RuntimeWarning)


class CouchbaseColumnarOfflineStoreConfig(FeastConfigBaseModel):
    """Offline store config for Couchbase Columnar"""

    type: Literal["couchbase.offline"] = "couchbase.offline"

    connection_string: Optional[StrictStr] = None
    user: Optional[StrictStr] = None
    password: Optional[StrictStr] = None
    timeout: StrictFloat = 120


class CouchbaseColumnarOfflineStore(OfflineStore):
    @staticmethod
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        """
        Fetch the latest rows for each join key.
        """
        warnings.warn(
            "This offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        assert isinstance(config.offline_store, CouchbaseColumnarOfflineStoreConfig)
        assert isinstance(data_source, CouchbaseColumnarSource)
        from_expression = data_source.get_table_query_string()

        partition_by_join_key_string = ", ".join(_append_alias(join_key_columns, "a"))
        if partition_by_join_key_string != "":
            partition_by_join_key_string = (
                "PARTITION BY " + partition_by_join_key_string
            )
        timestamps = [timestamp_field]
        if created_timestamp_column:
            timestamps.append(created_timestamp_column)
        timestamp_desc_string = " DESC, ".join(_append_alias(timestamps, "a")) + " DESC"
        a_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "a")
        )
        b_field_string = ", ".join(
            _append_alias(join_key_columns + feature_name_columns + timestamps, "b")
        )

        start_date_normalized = normalize_timestamp(start_date)
        end_date_normalized = normalize_timestamp(end_date)

        query = f"""
            SELECT
                {b_field_string}
                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
            FROM (
                SELECT {a_field_string},
                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS _feast_row
                FROM {from_expression} a
                WHERE a.{timestamp_field} BETWEEN '{start_date_normalized}' AND '{end_date_normalized}'
            ) b
            WHERE _feast_row = 1
            """

        return CouchbaseColumnarRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
            timestamp_field=timestamp_field,
        )

    @staticmethod
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: BaseRegistry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:
        """
        Retrieve historical features using point-in-time joins.
        """
        warnings.warn(
            "This offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        assert isinstance(config.offline_store, CouchbaseColumnarOfflineStoreConfig)
        for fv in feature_views:
            assert isinstance(fv.batch_source, CouchbaseColumnarSource)

        entity_schema = _get_entity_schema(entity_df, config)

        entity_df_event_timestamp_col = (
            offline_utils.infer_event_timestamp_from_entity_df(entity_schema)
        )

        entity_df_event_timestamp_range = _get_entity_df_event_timestamp_range(
            entity_df, entity_df_event_timestamp_col, config
        )

        @contextlib.contextmanager
        def query_generator() -> Iterator[str]:
            source = cast(CouchbaseColumnarSource, feature_views[0].batch_source)
            database = source.database
            scope = source.scope

            table_name = (
                f"{database}.{scope}.{offline_utils.get_temp_entity_table_name()}"
            )

            _upload_entity_df(config, entity_df, table_name)

            expected_join_keys = offline_utils.get_expected_join_keys(
                project, feature_views, registry
            )

            offline_utils.assert_expected_columns_in_entity_df(
                entity_schema, expected_join_keys, entity_df_event_timestamp_col
            )

            query_context = offline_utils.get_feature_view_query_context(
                feature_refs,
                feature_views,
                registry,
                project,
                entity_df_event_timestamp_range,
            )

            query_context_dict = [asdict(context) for context in query_context]

            try:
                query = build_point_in_time_query(
                    query_context_dict,
                    left_table_query_string=table_name,
                    entity_df_event_timestamp_col=entity_df_event_timestamp_col,
                    entity_df_columns=entity_schema.keys(),
                    query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                    full_feature_names=full_feature_names,
                )
                yield query
            finally:
                if table_name:
                    _execute_query(
                        config.offline_store,
                        f"DROP COLLECTION {table_name} IF EXISTS",
                    )

        return CouchbaseColumnarRetrievalJob(
            query=query_generator,
            config=config,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(entity_schema.keys() - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
            timestamp_field=entity_df_event_timestamp_col,
        )

    @staticmethod
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> RetrievalJob:
        """
        Fetch all rows from the specified table or query within the time range.
        """
        warnings.warn(
            "This offline store is an experimental feature in alpha development. "
            "Some functionality may still be unstable so functionality can change in the future.",
            RuntimeWarning,
        )
        assert isinstance(config.offline_store, CouchbaseColumnarOfflineStoreConfig)
        assert isinstance(data_source, CouchbaseColumnarSource)
        from_expression = data_source.get_table_query_string()

        timestamp_fields = [timestamp_field]
        if created_timestamp_column:
            timestamp_fields.append(created_timestamp_column)
        field_string = ", ".join(
            join_key_columns + feature_name_columns + timestamp_fields
        )
        start_date_normalized = (
            f"`{normalize_timestamp(start_date)}`" if start_date else None
        )
        end_date_normalized = f"`{normalize_timestamp(end_date)}`" if end_date else None
        timestamp_filter = get_timestamp_filter_sql(
            start_date_normalized,
            end_date_normalized,
            timestamp_field,
            cast_style="raw",
            quote_fields=False,
        )

        query = f"""
        SELECT {field_string}
        FROM {from_expression}
        WHERE {timestamp_filter}
        """

        return CouchbaseColumnarRetrievalJob(
            query=query,
            config=config,
            full_feature_names=False,
            on_demand_feature_views=None,
            timestamp_field=timestamp_field,
        )


class CouchbaseColumnarRetrievalJob(RetrievalJob):
    def __init__(
        self,
        query: Union[str, Callable[[], ContextManager[str]]],
        config: RepoConfig,
        full_feature_names: bool,
        timestamp_field: str,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        if not isinstance(query, str):
            self._query_generator = query
        else:

            @contextlib.contextmanager
            def query_generator() -> Iterator[str]:
                assert isinstance(query, str)
                yield query

            self._query_generator = query_generator
        self._config = config
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = on_demand_feature_views or []
        self._metadata = metadata
        self._timestamp_field = timestamp_field

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> List[OnDemandFeatureView]:
        return self._on_demand_feature_views

    def _to_df_internal(self, timeout: Optional[int] = None) -> pd.DataFrame:
        # Use PyArrow to convert the result to a pandas DataFrame
        return self._to_arrow_internal(timeout).to_pandas()

    def to_sql(self) -> str:
        with self._query_generator() as query:
            return query

    def _to_arrow_internal(self, timeout: Optional[int] = None) -> pa.Table:
        with self._query_generator() as query:
            res = _execute_query(self._config.offline_store, query)
            rows = res.get_all_rows()

            processed_rows = []
            for row in rows:
                processed_row = {}
                for key, value in row.items():
                    if key == self._timestamp_field and value is not None:
                        # Parse and ensure timezone-aware datetime
                        processed_row[key] = pd.to_datetime(value, utc=True)
                    else:
                        processed_row[key] = np.nan if value is None else value
                processed_rows.append(processed_row)

            # Convert to PyArrow table
            table = pa.Table.from_pylist(processed_rows)
            return table

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata

    def persist(
        self,
        storage: SavedDatasetStorage,
        allow_overwrite: Optional[bool] = False,
        timeout: Optional[int] = None,
    ):
        assert isinstance(storage, SavedDatasetCouchbaseColumnarStorage)
        table_name = f"{storage.couchbase_options._database}.{storage.couchbase_options._scope}.{offline_utils.get_temp_entity_table_name()}"
        df_to_columnar(self.to_df(), table_name, self._config.offline_store)


def _get_columnar_cluster(config: CouchbaseColumnarOfflineStoreConfig) -> Cluster:
    assert config.connection_string is not None
    assert config.user is not None
    assert config.password is not None

    cred = Credential.from_username_and_password(config.user, config.password)
    timeout_opts = TimeoutOptions(dispatch_timeout=timedelta(seconds=120))
    return Cluster.create_instance(
        config.connection_string, cred, ClusterOptions(timeout_options=timeout_opts)
    )


def _execute_query(
    config: CouchbaseColumnarOfflineStoreConfig,
    query: str,
    named_params: Optional[Dict[str, Any]] = None,
) -> BlockingQueryResult:
    cluster = _get_columnar_cluster(config)
    return cluster.execute_query(
        query,
        QueryOptions(
            named_parameters=named_params, timeout=timedelta(seconds=config.timeout)
        ),
    )


def df_to_columnar(
    df: pd.DataFrame,
    table_name: str,
    offline_store: CouchbaseColumnarOfflineStoreConfig,
):
    df_copy = df.copy()
    insert_values = df_copy.apply(
        lambda row: {
            col: (
                normalize_timestamp(row[col], "%Y-%m-%dT%H:%M:%S.%f+00:00")
                if isinstance(row[col], pd.Timestamp)
                else row[col]
            )
            for col in df_copy.columns
        },
        axis=1,
    ).tolist()

    create_collection_query = f"CREATE COLLECTION {table_name} IF NOT EXISTS PRIMARY KEY(pk: UUID) AUTOGENERATED;"
    insert_query = f"INSERT INTO {table_name} ({insert_values});"

    _execute_query(offline_store, create_collection_query)
    _execute_query(offline_store, insert_query)


def _upload_entity_df(
    config: RepoConfig, entity_df: Union[pd.DataFrame, str], table_name: str
):
    if isinstance(entity_df, pd.DataFrame):
        df_to_columnar(entity_df, table_name, config.offline_store)
    elif isinstance(entity_df, str):
        # If the entity_df is a string (SQL query), create a Columnar collection out of it
        create_collection_query = f"""
        CREATE COLLECTION {table_name} IF NOT EXISTS
        PRIMARY KEY(pk: UUID) AUTOGENERATED
        AS {entity_df}
        """
        _execute_query(config.offline_store, create_collection_query)
    else:
        raise InvalidEntityType(type(entity_df))


def _get_entity_df_event_timestamp_range(
    entity_df: Union[pd.DataFrame, str],
    entity_df_event_timestamp_col: str,
    config: RepoConfig,
) -> Tuple[datetime, datetime]:
    if isinstance(entity_df, pd.DataFrame):
        entity_df_event_timestamp = entity_df.loc[
            :, entity_df_event_timestamp_col
        ].infer_objects()
        if pd.api.types.is_string_dtype(entity_df_event_timestamp):
            entity_df_event_timestamp = pd.to_datetime(
                entity_df_event_timestamp, utc=True
            )
        entity_df_event_timestamp_range = (
            entity_df_event_timestamp.min().to_pydatetime(),
            entity_df_event_timestamp.max().to_pydatetime(),
        )

    elif isinstance(entity_df, str):
        query = f"""
            SELECT
                MIN({entity_df_event_timestamp_col}) AS min,
                MAX({entity_df_event_timestamp_col}) AS max
            FROM ({entity_df}) AS tmp_alias
            """

        res = _execute_query(config.offline_store, query)
        rows = res.get_all_rows()

        if not rows:
            raise ZeroRowsQueryResult(query)

        # Convert the string timestamps to datetime objects
        min_ts = pd.to_datetime(rows[0]["min"], utc=True).to_pydatetime()
        max_ts = pd.to_datetime(rows[0]["max"], utc=True).to_pydatetime()
        entity_df_event_timestamp_range = (min_ts, max_ts)
    else:
        raise InvalidEntityType(type(entity_df))
    return entity_df_event_timestamp_range


def _escape_column(column: str) -> str:
    """Wrap column names in backticks to handle reserved words."""
    return f"`{column}`"


def _append_alias(field_names: List[str], alias: str) -> List[str]:
    """Append alias to escaped column names."""
    return [f"{alias}.{_escape_column(field_name)}" for field_name in field_names]


def build_point_in_time_query(
    feature_view_query_contexts: List[dict],
    left_table_query_string: str,
    entity_df_event_timestamp_col: str,
    entity_df_columns: KeysView[str],
    query_template: str,
    full_feature_names: bool = False,
) -> str:
    """Build point-in-time query between each feature view table and the entity dataframe for Couchbase Columnar"""
    template = Environment(loader=BaseLoader()).from_string(source=query_template)
    final_output_feature_names = list(entity_df_columns)
    final_output_feature_names.extend(
        [
            (
                f"{fv['name']}__{fv['field_mapping'].get(feature, feature)}"
                if full_feature_names
                else fv["field_mapping"].get(feature, feature)
            )
            for fv in feature_view_query_contexts
            for feature in fv["features"]
        ]
    )

    # Add additional fields to dict
    template_context = {
        "left_table_query_string": left_table_query_string,
        "entity_df_event_timestamp_col": entity_df_event_timestamp_col,
        "unique_entity_keys": set(
            [entity for fv in feature_view_query_contexts for entity in fv["entities"]]
        ),
        "featureviews": feature_view_query_contexts,
        "full_feature_names": full_feature_names,
        "final_output_feature_names": final_output_feature_names,
    }

    query = template.render(template_context)
    return query


def get_couchbase_query_schema(config, entity_df: str) -> Dict[str, np.dtype]:
    df_query = f"({entity_df}) AS sub"
    res = _execute_query(config.offline_store, f"SELECT sub.* FROM {df_query} LIMIT 1")
    rows = res.get_all_rows()

    if rows and len(rows) > 0:
        # Get the first row
        first_row = rows[0]
        # Create dictionary mapping each column to dtype('O')
        return {key: np.dtype("O") for key in first_row.keys()}

    return {}


def _get_entity_schema(
    entity_df: Union[pd.DataFrame, str],
    config: RepoConfig,
) -> Dict[str, np.dtype]:
    if isinstance(entity_df, pd.DataFrame):
        return dict(zip(entity_df.columns, entity_df.dtypes))

    elif isinstance(entity_df, str):
        return get_couchbase_query_schema(config, entity_df)
    else:
        raise InvalidEntityType(type(entity_df))


MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN = """
WITH entity_dataframe AS (
    SELECT e.*,
        e.`{{entity_df_event_timestamp_col}}` AS entity_timestamp
        {% for featureview in featureviews -%}
            {% if featureview.entities -%}
            ,CONCAT(
                {% for entity in featureview.entities -%}
                    TOSTRING(e.`{{entity}}`),
                {% endfor -%}
                TOSTRING(e.`{{entity_df_event_timestamp_col}}`)
            ) AS `{{featureview.name}}__entity_row_unique_id`
            {% else -%}
            ,TOSTRING(e.`{{entity_df_event_timestamp_col}}`) AS `{{featureview.name}}__entity_row_unique_id`
            {% endif -%}
        {% endfor %}
    FROM {{ left_table_query_string }} e
),

{% for featureview in featureviews %}

`{{ featureview.name }}__entity_dataframe` AS (
    SELECT
        {% if featureview.entities %}`{{ featureview.entities | join('`, `') }}`,{% endif %}
        entity_timestamp,
        `{{featureview.name}}__entity_row_unique_id`
    FROM entity_dataframe
    GROUP BY
        {% if featureview.entities %}`{{ featureview.entities | join('`, `')}}`,{% endif %}
        entity_timestamp,
        `{{featureview.name}}__entity_row_unique_id`
),

/*
 This query template performs the point-in-time correctness join for a single feature set table
 to the provided entity table.

 1. We first join the current feature_view to the entity dataframe that has been passed.
 This JOIN has the following logic:
    - For each row of the entity dataframe, only keep the rows where the `timestamp_field`
    is less than the one provided in the entity dataframe
    - If there a TTL for the current feature_view, also keep the rows where the `timestamp_field`
    is higher the the one provided minus the TTL
    - For each row, Join on the entity key and retrieve the `entity_row_unique_id` that has been
    computed previously

 The output of this CTE will contain all the necessary information and already filtered out most
 of the data that is not relevant.
*/
`{{ featureview.name }}__subquery` AS (
    LET max_ts = (SELECT RAW MAX(entity_timestamp) FROM entity_dataframe)[0]
    SELECT s.* FROM (
        LET min_ts = (SELECT RAW MIN(entity_timestamp) FROM entity_dataframe)[0]
        SELECT
            `{{ featureview.timestamp_field }}` as event_timestamp,
            {{ '`' ~ featureview.created_timestamp_column ~ '` as created_timestamp,' if featureview.created_timestamp_column else '' }}
            {{ featureview.entity_selections | join(', ')}}{% if featureview.entity_selections %},{% else %}{% endif %}
            {% for feature in featureview.features -%}
                `{{ feature }}` as {% if full_feature_names %}`{{ featureview.name }}__{{featureview.field_mapping.get(feature, feature)}}`{% else %}`{{ featureview.field_mapping.get(feature, feature) }}`{% endif %}{% if not loop.last %}, {% endif %}
            {%- endfor %}
        FROM {{ featureview.table_subquery }} AS sub
        WHERE `{{ featureview.timestamp_field }}` <= max_ts
        {% if featureview.ttl == 0 %}{% else %}
        AND date_diff_str(min_ts, `{{ featureview.timestamp_field }}`, "second") <= {{ featureview.ttl }}
        {% endif %}
    ) s
),

`{{ featureview.name }}__base` AS (
    SELECT
        subquery.*,
        entity_dataframe.entity_timestamp,
        entity_dataframe.`{{featureview.name}}__entity_row_unique_id`
    FROM `{{ featureview.name }}__subquery` AS subquery
    INNER JOIN `{{ featureview.name }}__entity_dataframe` AS entity_dataframe
    ON TRUE
        AND subquery.event_timestamp <= entity_dataframe.entity_timestamp
        {% if featureview.ttl == 0 %}{% else %}
        AND date_diff_str(entity_dataframe.entity_timestamp, subquery.event_timestamp, "second") <= {{ featureview.ttl }}
        {% endif %}
        {% for entity in featureview.entities %}
        AND subquery.`{{ entity }}` = entity_dataframe.`{{ entity }}`
        {% endfor %}
),

/*
 2. If the `created_timestamp_column` has been set, we need to
 deduplicate the data first. This is done by calculating the
 `MAX(created_at_timestamp)` for each event_timestamp.
 We then join the data on the next CTE
*/
{% if featureview.created_timestamp_column %}
`{{ featureview.name }}__dedup` AS (
    SELECT
        `{{featureview.name}}__entity_row_unique_id`,
        event_timestamp,
        MAX(created_timestamp) AS created_timestamp
    FROM `{{ featureview.name }}__base`
    GROUP BY `{{featureview.name}}__entity_row_unique_id`, event_timestamp
),
{% endif %}

/*
 3. The data has been filtered during the first CTE "*__base"
 Thus we only need to compute the latest timestamp of each feature.
*/
`{{ featureview.name }}__latest` AS (
    SELECT
        event_timestamp
        {% if featureview.created_timestamp_column %},created_timestamp{% endif %},
        `{{featureview.name}}__entity_row_unique_id`
    FROM (
        SELECT base.*,
            ROW_NUMBER() OVER(
                PARTITION BY base.`{{featureview.name}}__entity_row_unique_id`
                ORDER BY base.event_timestamp DESC
                {% if featureview.created_timestamp_column %}, base.created_timestamp DESC{% endif %}
            ) AS row_number
        FROM `{{ featureview.name }}__base` base
        {% if featureview.created_timestamp_column %}
        INNER JOIN `{{ featureview.name }}__dedup` dedup
        ON base.`{{featureview.name}}__entity_row_unique_id` = dedup.`{{featureview.name}}__entity_row_unique_id`
        AND base.event_timestamp = dedup.event_timestamp
        AND base.created_timestamp = dedup.created_timestamp
        {% endif %}
    ) AS sub
    WHERE sub.row_number = 1
),

/*
 4. Once we know the latest value of each feature for a given timestamp,
 we can join again the data back to the original "base" dataset
*/
`{{ featureview.name }}__cleaned` AS (
    SELECT base.*
    FROM `{{ featureview.name }}__base` AS base
    INNER JOIN `{{ featureview.name }}__latest` AS latest
    ON base.`{{featureview.name}}__entity_row_unique_id` = latest.`{{featureview.name}}__entity_row_unique_id`
    AND base.event_timestamp = latest.event_timestamp
    {% if featureview.created_timestamp_column %}
    AND base.created_timestamp = latest.created_timestamp
    {% endif %}
){% if not loop.last %},{% endif %}

{% endfor %}

/*
 Joins the outputs of multiple time travel joins to a single table.
 The entity_dataframe dataset being our source of truth here.
 */
SELECT DISTINCT
    {%- set fields = [] %}
    {%- for feature_name in final_output_feature_names %}
    {%- if '__' not in feature_name %}
        {%- set ns = namespace(found=false) %}
        {%- for fv in featureviews %}
            {%- for feature in fv.features %}
                {%- if feature == feature_name %}
                    {%- set ns.found = true %}
                    {%- if full_feature_names %}
                        {%- set _ = fields.append('IFMISSINGORNULL(`' ~ fv.name ~ '_final`.`' ~ fv.name ~ '__' ~ feature ~ '`, null) AS `' ~ fv.name ~ '__' ~ feature ~ '`') %}
                    {%- else %}
                        {%- set _ = fields.append('IFMISSINGORNULL(`' ~ fv.name ~ '_final`.`' ~ feature ~ '`, null) AS `' ~ feature ~ '`') %}
                    {%- endif %}
                {%- endif %}
            {%- endfor %}
        {%- endfor %}
        {%- if not ns.found %}
            {%- if feature_name == 'feature_name' %}
                {%- set _ = fields.append('IFMISSINGORNULL(`field_mapping_final`.`' ~ feature_name ~ '`, null) AS `' ~ feature_name ~ '`') %}
            {%- else %}
                {%- set _ = fields.append('main_entity.`' ~ feature_name ~ '`') %}
            {%- endif %}
        {%- endif %}
    {%- else %}
        {%- set feature_parts = feature_name.split('__') %}
        {%- set fv_name = feature_parts[0] %}
        {%- set feature = feature_parts[1] %}
        {%- if feature_name == 'field_mapping__feature_name' %}
            {%- set _ = fields.append('IFMISSINGORNULL(`field_mapping_final`.`field_mapping__feature_name`, null) AS `field_mapping__feature_name`') %}
        {%- else %}
            {%- set _ = fields.append('IFMISSINGORNULL(`' ~ fv_name ~ '_final`.`' ~ feature_name ~ '`, null) AS `' ~ feature_name ~ '`') %}
        {%- endif %}
    {%- endif %}
    {%- endfor %}
    {{ fields | reject('none') | join(',\n    ') }}
FROM entity_dataframe AS main_entity

{%- for featureview in featureviews %}
LEFT JOIN (
    SELECT
        `{{featureview.name}}__entity_row_unique_id`,
        {% for feature in featureview.features -%}
        IFMISSINGORNULL(`{% if full_feature_names %}{{ featureview.name }}__{{ featureview.field_mapping.get(feature, feature) }}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}`, null) AS `{% if full_feature_names %}{{ featureview.name }}__{{ featureview.field_mapping.get(feature, feature) }}{% else %}{{ featureview.field_mapping.get(feature, feature) }}{% endif %}`{% if not loop.last %},{% endif %}
        {% endfor %}
    FROM `{{ featureview.name }}__cleaned`
) AS `{{featureview.name}}_final`
ON main_entity.`{{featureview.name}}__entity_row_unique_id` = `{{featureview.name}}_final`.`{{featureview.name}}__entity_row_unique_id`
{% endfor %}
"""



from datetime import datetime, timedelta
from typing import List, Optional, Union

import pyarrow as pa

from feast import BatchFeatureView, StreamFeatureView
from feast.data_source import DataSource
from feast.infra.compute_engines.dag.context import ColumnInfo, ExecutionContext
from feast.infra.compute_engines.dag.model import DAGFormat
from feast.infra.compute_engines.dag.node import DAGNode
from feast.infra.compute_engines.local.arrow_table_value import ArrowTableValue
from feast.infra.compute_engines.local.backends.base import DataFrameBackend
from feast.infra.compute_engines.local.local_node import LocalNode
from feast.infra.compute_engines.utils import (
    create_offline_store_retrieval_job,
)
from feast.infra.offline_stores.offline_utils import (
    infer_event_timestamp_from_entity_df,
)
from feast.utils import _convert_arrow_to_proto

ENTITY_TS_ALIAS = "__entity_event_timestamp"


class LocalSourceReadNode(LocalNode):
    def __init__(
        self,
        name: str,
        source: DataSource,
        column_info: ColumnInfo,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ):
        super().__init__(name)
        self.source = source
        self.column_info = column_info
        self.start_time = start_time
        self.end_time = end_time

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        retrieval_job = create_offline_store_retrieval_job(
            data_source=self.source,
            context=context,
            start_time=self.start_time,
            end_time=self.end_time,
            column_info=self.column_info,
        )
        arrow_table = retrieval_job.to_arrow()
        if self.column_info.field_mapping:
            arrow_table = arrow_table.rename_columns(
                [
                    self.column_info.field_mapping.get(col, col)
                    for col in arrow_table.column_names
                ]
            )
        return ArrowTableValue(data=arrow_table)


class LocalJoinNode(LocalNode):
    def __init__(
        self,
        name: str,
        column_info: ColumnInfo,
        backend: DataFrameBackend,
        inputs: Optional[List["DAGNode"]] = None,
        how: str = "inner",
    ):
        super().__init__(name, inputs or [])
        self.column_info = column_info
        self.backend = backend
        self.how = how

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_values = self.get_input_values(context)
        for val in input_values:
            val.assert_format(DAGFormat.ARROW)

        # Convert all upstream ArrowTables to backend DataFrames
        joined_df = self.backend.from_arrow(input_values[0].data)
        for val in input_values[1:]:
            next_df = self.backend.from_arrow(val.data)
            joined_df = self.backend.join(
                joined_df,
                next_df,
                on=self.column_info.join_keys,
                how=self.how,
            )

        # If entity_df is provided, join it in last
        if context.entity_df is not None:
            entity_df = self.backend.from_arrow(pa.Table.from_pandas(context.entity_df))

            entity_schema = dict(zip(entity_df.columns, entity_df.dtypes))
            entity_ts_col = infer_event_timestamp_from_entity_df(entity_schema)

            if entity_ts_col != ENTITY_TS_ALIAS:
                entity_df = self.backend.rename_columns(
                    entity_df, {entity_ts_col: ENTITY_TS_ALIAS}
                )

            joined_df = self.backend.join(
                entity_df,
                joined_df,
                on=self.column_info.join_keys,
                how="left",
            )

        result = self.backend.to_arrow(joined_df)
        return ArrowTableValue(result)


class LocalFilterNode(LocalNode):
    def __init__(
        self,
        name: str,
        column_info: ColumnInfo,
        backend: DataFrameBackend,
        filter_expr: Optional[str] = None,
        ttl: Optional[timedelta] = None,
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.column_info = column_info
        self.backend = backend
        self.filter_expr = filter_expr
        self.ttl = ttl

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        df = self.backend.from_arrow(input_table)

        timestamp_column = self.column_info.timestamp_column

        if ENTITY_TS_ALIAS in self.backend.columns(df):
            # filter where feature.ts <= entity.event_timestamp
            df = df[df[timestamp_column] <= df[ENTITY_TS_ALIAS]]

            # TTL: feature.ts >= entity.event_timestamp - ttl
            if self.ttl:
                lower_bound = df[ENTITY_TS_ALIAS] - self.backend.to_timedelta_value(
                    self.ttl
                )
                df = df[df[timestamp_column] >= lower_bound]

        # Optional user-defined filter expression (e.g., "value > 0")
        if self.filter_expr:
            df = self.backend.filter(df, self.filter_expr)

        result = self.backend.to_arrow(df)
        output = ArrowTableValue(result)
        context.node_outputs[self.name] = output
        return output


class LocalAggregationNode(LocalNode):
    def __init__(
        self,
        name: str,
        backend: DataFrameBackend,
        group_keys: list[str],
        agg_ops: dict,
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.backend = backend
        self.group_keys = group_keys
        self.agg_ops = agg_ops

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        df = self.backend.from_arrow(input_table)
        grouped_df = self.backend.groupby_agg(df, self.group_keys, self.agg_ops)
        result = self.backend.to_arrow(grouped_df)
        output = ArrowTableValue(result)
        return output


class LocalDedupNode(LocalNode):
    def __init__(
        self, name: str, column_info: ColumnInfo, backend: DataFrameBackend, inputs=None
    ):
        super().__init__(name, inputs=inputs)
        self.column_info = column_info
        self.backend = backend

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        df = self.backend.from_arrow(input_table)

        # Extract join_keys, timestamp, and created_ts from context

        # Dedup strategy: sort and drop_duplicates
        dedup_keys = self.column_info.join_keys
        if dedup_keys:
            sort_keys = [self.column_info.timestamp_column]
            if (
                self.column_info.created_timestamp_column
                and self.column_info.created_timestamp_column in df.columns
            ):
                sort_keys.append(self.column_info.created_timestamp_column)

            df = self.backend.drop_duplicates(
                df, keys=dedup_keys, sort_by=sort_keys, ascending=False
            )
        result = self.backend.to_arrow(df)
        output = ArrowTableValue(result)
        context.node_outputs[self.name] = output
        return output


class LocalTransformationNode(LocalNode):
    def __init__(
        self, name: str, transformation_fn, backend: DataFrameBackend, inputs=None
    ):
        super().__init__(name, inputs=inputs)
        self.transformation_fn = transformation_fn
        self.backend = backend

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        df = self.backend.from_arrow(input_table)
        transformed_df = self.transformation_fn(df)
        result = self.backend.to_arrow(transformed_df)
        output = ArrowTableValue(result)
        context.node_outputs[self.name] = output
        return output


class LocalValidationNode(LocalNode):
    def __init__(
        self, name: str, validation_config, backend: DataFrameBackend, inputs=None
    ):
        super().__init__(name, inputs=inputs)
        self.validation_config = validation_config
        self.backend = backend

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        df = self.backend.from_arrow(input_table)
        # Placeholder for actual validation logic
        if self.validation_config:
            print(f"[Validation: {self.name}] Passed.")
        result = self.backend.to_arrow(df)
        output = ArrowTableValue(result)
        context.node_outputs[self.name] = output
        return output


class LocalOutputNode(LocalNode):
    def __init__(
        self,
        name: str,
        feature_view: Union[BatchFeatureView, StreamFeatureView],
        inputs=None,
    ):
        super().__init__(name, inputs=inputs)
        self.feature_view = feature_view

    def execute(self, context: ExecutionContext) -> ArrowTableValue:
        input_table = self.get_single_table(context).data
        context.node_outputs[self.name] = input_table

        if input_table.num_rows == 0:
            return input_table

        if self.feature_view.online:
            online_store = context.online_store

            join_key_to_value_type = {
                entity.name: entity.dtype.to_value_type()
                for entity in self.feature_view.entity_columns
            }

            rows_to_write = _convert_arrow_to_proto(
                input_table, self.feature_view, join_key_to_value_type
            )

            online_store.online_write_batch(
                config=context.repo_config,
                table=self.feature_view,
                data=rows_to_write,
                progress=lambda x: None,
            )

        if self.feature_view.offline:
            offline_store = context.offline_store
            offline_store.offline_write_batch(
                config=context.repo_config,
                feature_view=self.feature_view,
                table=input_table,
                progress=lambda x: None,
            )

        return input_table



import click
import yaml

from feast import utils
from feast.cli.cli_options import tagsOption
from feast.errors import FeastObjectNotFoundException
from feast.repo_operations import create_feature_store


@click.group(name="on-demand-feature-views")
def on_demand_feature_views_cmd():
    """
    [Experimental] Access on demand feature views
    """
    pass


@on_demand_feature_views_cmd.command("describe")
@click.argument("name", type=click.STRING)
@click.pass_context
def on_demand_feature_view_describe(ctx: click.Context, name: str):
    """
    [Experimental] Describe an on demand feature view
    """
    store = create_feature_store(ctx)

    try:
        on_demand_feature_view = store.get_on_demand_feature_view(name)
    except FeastObjectNotFoundException as e:
        print(e)
        exit(1)

    print(
        yaml.dump(
            yaml.safe_load(str(on_demand_feature_view)),
            default_flow_style=False,
            sort_keys=False,
        )
    )


@on_demand_feature_views_cmd.command(name="list")
@tagsOption
@click.pass_context
def on_demand_feature_view_list(ctx: click.Context, tags: list[str]):
    """
    [Experimental] List all on demand feature views
    """
    store = create_feature_store(ctx)
    table = []
    tags_filter = utils.tags_list_to_dict(tags)
    for on_demand_feature_view in store.list_on_demand_feature_views(tags=tags_filter):
        table.append([on_demand_feature_view.name])

    from tabulate import tabulate

    print(tabulate(table, headers=["NAME"], tablefmt="plain"))



import os
import tempfile
from typing import Any, Dict, Optional

from feast.data_format import ParquetFormat
from feast.data_source import DataSource
from feast.feature_logging import LoggingDestination
from feast.infra.offline_stores.contrib.ray_offline_store.ray import (
    RayOfflineStoreConfig,
)
from feast.infra.offline_stores.file_source import (
    FileLoggingDestination,
    FileSource,
    SavedDatasetFileStorage,
)
from feast.repo_config import FeastConfigBaseModel
from feast.saved_dataset import SavedDatasetStorage
from tests.integration.feature_repos.integration_test_repo_config import (
    IntegrationTestRepoConfig,
)
from tests.integration.feature_repos.universal.data_source_creator import (
    DataSourceCreator,
)


class RayDataSourceCreator(DataSourceCreator):
    def __init__(self, project_name: str, *args, **kwargs):
        super().__init__(project_name, *args, **kwargs)
        self.offline_store_config = RayOfflineStoreConfig(
            type="ray",
            storage_path="/tmp/ray-storage",
            ray_address=None,
            broadcast_join_threshold_mb=25,
            max_parallelism_multiplier=1,
            target_partition_size_mb=16,
            enable_ray_logging=False,
            ray_conf={
                "num_cpus": 1,
                "object_store_memory": 80 * 1024 * 1024,
                "_memory": 400 * 1024 * 1024,
            },
        )
        self.files: list[Any] = []
        self.dirs: list[str] = []

    def create_offline_store_config(self) -> FeastConfigBaseModel:
        return self.offline_store_config

    def create_data_source(
        self,
        df: Any,
        destination_name: str,
        created_timestamp_column: Optional[Any] = "created_ts",
        field_mapping: Optional[Dict[str, str]] = None,
        timestamp_field: Optional[str] = "ts",
    ) -> DataSource:
        # For Ray, we'll use parquet files as the underlying storage
        destination_name = self.get_prefixed_table_name(destination_name)

        f = tempfile.NamedTemporaryFile(
            prefix=f"{self.project_name}_{destination_name}",
            suffix=".parquet",
            delete=False,
        )
        df.to_parquet(f.name)
        self.files.append(f)

        return FileSource(
            file_format=ParquetFormat(),
            path=f.name,
            timestamp_field=timestamp_field,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping or {"ts_1": "ts"},
        )

    def get_prefixed_table_name(self, suffix: str) -> str:
        return f"{self.project_name}.{suffix}"

    def create_saved_dataset_destination(self) -> SavedDatasetStorage:
        d = tempfile.mkdtemp(prefix=self.project_name)
        self.dirs.append(d)
        return SavedDatasetFileStorage(
            path=d,
            file_format=ParquetFormat(),
        )

    def create_logged_features_destination(self) -> LoggingDestination:
        d = tempfile.mkdtemp(prefix=self.project_name)
        self.dirs.append(d)
        return FileLoggingDestination(path=d)

    def teardown(self) -> None:
        # Clean up any temporary files or resources
        import shutil

        for f in self.files:
            f.close()
            try:
                os.unlink(f.name)
            except OSError:
                pass

        for d in self.dirs:
            if os.path.exists(d):
                shutil.rmtree(d)

    def get_saved_dataset_data_source(self) -> Dict[str, str]:
        return {
            "type": "parquet",
            "path": "data/saved_dataset.parquet",
        }

    @staticmethod
    def xdist_groups() -> list[str]:
        """
        Return xdist group names for Ray tests.
        This ensures all Ray tests run on the same pytest worker to avoid OOM issues.
        """
        return ["ray"]


# Define the full repo configurations for Ray offline store
FULL_REPO_CONFIGS = [
    IntegrationTestRepoConfig(
        provider="local",
        offline_store_creator=RayDataSourceCreator,
    ),
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/types/Field.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from feast.protos.feast.types import Value_pb2 as feast_dot_types_dot_Value__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x66\x65\x61st/types/Field.proto\x12\x0b\x66\x65\x61st.types\x1a\x17\x66\x65\x61st/types/Value.proto\"\xaf\x01\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x05value\x18\x02 \x01(\x0e\x32\x1b.feast.types.ValueType.Enum\x12*\n\x04tags\x18\x03 \x03(\x0b\x32\x1c.feast.types.Field.TagsEntry\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42Q\n\x11\x66\x65\x61st.proto.typesB\nFieldProtoZ0github.com/feast-dev/feast/go/protos/feast/typesb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.types.Field_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\021feast.proto.typesB\nFieldProtoZ0github.com/feast-dev/feast/go/protos/feast/types'
  _globals['_FIELD_TAGSENTRY']._options = None
  _globals['_FIELD_TAGSENTRY']._serialized_options = b'8\001'
  _globals['_FIELD']._serialized_start=66
  _globals['_FIELD']._serialized_end=241
  _globals['_FIELD_TAGSENTRY']._serialized_start=198
  _globals['_FIELD_TAGSENTRY']._serialized_end=241
# @@protoc_insertion_point(module_scope)



from __future__ import absolute_import

import base64
import json
import logging
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

from elasticsearch import Elasticsearch, helpers

from feast import Entity, FeatureView, RepoConfig
from feast.infra.key_encoding_utils import (
    deserialize_entity_key,
    get_list_val_str,
    serialize_entity_key,
)
from feast.infra.online_stores.online_store import OnlineStore
from feast.infra.online_stores.vector_store import VectorStoreConfig
from feast.protos.feast.types.EntityKey_pb2 import EntityKey as EntityKeyProto
from feast.protos.feast.types.Value_pb2 import Value as ValueProto
from feast.repo_config import FeastConfigBaseModel
from feast.utils import (
    _build_retrieve_online_document_record,
    _get_feature_view_vector_field_metadata,
    to_naive_utc,
)


class ElasticSearchOnlineStoreConfig(FeastConfigBaseModel, VectorStoreConfig):
    """
    Configuration for the ElasticSearch online store.
    NOTE: The class *must* end with the `OnlineStoreConfig` suffix.
    """

    type: str = "elasticsearch"

    host: Optional[str] = None
    user: Optional[str] = None
    password: Optional[str] = None
    port: Optional[int] = None
    index: Optional[str] = None
    scheme: Optional[str] = "http"
    vector_field_path: Optional[str] = "embedding.vector_value"

    # The number of rows to write in a single batch
    write_batch_size: Optional[int] = 40


class ElasticSearchOnlineStore(OnlineStore):
    _client: Optional[Elasticsearch] = None

    def _get_client(self, config: RepoConfig) -> Elasticsearch:
        online_store_config = config.online_store
        assert isinstance(online_store_config, ElasticSearchOnlineStoreConfig)

        user = online_store_config.user if online_store_config.user is not None else ""
        password = (
            online_store_config.password
            if online_store_config.password is not None
            else ""
        )

        if self._client:
            return self._client
        else:
            self._client = Elasticsearch(
                hosts=[
                    {
                        "host": online_store_config.host or "localhost",
                        "port": online_store_config.port or 9200,
                        "scheme": online_store_config.scheme or "http",
                    }
                ],
                basic_auth=(user, password),
            )
            return self._client

    def _bulk_batch_actions(self, table: FeatureView, batch: List[Dict[str, Any]]):
        for row in batch:
            yield {
                "_index": table.name,
                "_id": f"{row['entity_key']}_{row['timestamp']}",
                "_source": row,
            }

    def online_write_batch(
        self,
        config: RepoConfig,
        table: FeatureView,
        data: List[
            Tuple[EntityKeyProto, Dict[str, ValueProto], datetime, Optional[datetime]]
        ],
        progress: Optional[Callable[[int], Any]],
    ) -> None:
        insert_values = []
        grouped_docs: dict[str, dict[str, Any]] = defaultdict(
            lambda: {
                "features": {},
                "vector_value": None,
                "timestamp": None,
                "created_ts": None,
            }
        )
        for entity_key, values, timestamp, created_ts in data:
            entity_key_bin = serialize_entity_key(
                entity_key,
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            encoded_entity_key = base64.b64encode(entity_key_bin).decode("utf-8")
            timestamp = to_naive_utc(timestamp)
            if created_ts is not None:
                created_ts = to_naive_utc(created_ts)

            doc_key = f"{encoded_entity_key}_{timestamp}"

            for feature_name, value in values.items():
                doc = _encode_feature_value(value)
                grouped_docs[doc_key]["features"][feature_name] = doc
                grouped_docs[doc_key]["timestamp"] = timestamp
                grouped_docs[doc_key]["created_ts"] = created_ts
                grouped_docs[doc_key]["entity_key"] = encoded_entity_key

                insert_values = [
                    {
                        "entity_key": document["entity_key"],
                        "timestamp": document["timestamp"],
                        "created_ts": document["created_ts"],
                        **(document["features"] or {}),
                    }
                    for document in grouped_docs.values()
                ]

        batch_size = config.online_store.write_batch_size
        for i in range(0, len(insert_values), batch_size):
            batch = insert_values[i : i + batch_size]
            actions = self._bulk_batch_actions(table, batch)
            helpers.bulk(self._get_client(config), actions, refresh="wait_for")

    def online_read(
        self,
        config: RepoConfig,
        table: FeatureView,
        entity_keys: List[EntityKeyProto],
        requested_features: Optional[List[str]] = None,
    ) -> List[Tuple[Optional[datetime], Optional[Dict[str, ValueProto]]]]:
        encoded_entity_keys = [
            base64.b64encode(
                serialize_entity_key(ek, config.entity_key_serialization_version)
            ).decode("utf-8")
            for ek in entity_keys
        ]

        # Determine which fields to retrieve
        includes = ["timestamp", "created_ts", "entity_key"]
        if requested_features:
            includes.extend(requested_features)
        else:
            includes.append("*")

        body = {
            "_source": {"includes": includes, "excludes": ["*.vector_value"]},
            "query": {
                "bool": {"filter": [{"terms": {"entity_key": encoded_entity_keys}}]}
            },
        }

        response = self._get_client(config).search(index=table.name, body=body)

        results = []

        for hit in response["hits"]["hits"]:
            source = hit["_source"]
            timestamp = source.get("timestamp")
            timestamp = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")

            features: Dict[str, ValueProto] = {}

            fields_to_extract = (
                requested_features if requested_features else source.keys()
            )

            for feature_name in fields_to_extract:
                if feature_name in ("timestamp", "created_ts", "entity_key"):
                    continue
                feature_obj = source.get(feature_name)
                if not feature_obj or "feature_value" not in feature_obj:
                    continue
                try:
                    features[feature_name] = _to_value_proto(feature_obj)
                except Exception as e:
                    raise ValueError(
                        f"Failed to parse feature '{feature_name}' from hit: {e}"
                    )

            results.append((timestamp, features if features else None))

        return results

    def create_index(self, config: RepoConfig, table: FeatureView):
        """
        Create an index in ElasticSearch for the given table.
        TODO: This method can be exposed to users to customize the indexing functionality.
        Args:
            config: Feast repo configuration object.
            table: FeatureView table for which the index needs to be created.
        """
        vector_field_length = getattr(
            _get_feature_view_vector_field_metadata(table), "vector_length", 512
        )

        index_mapping = {
            "dynamic_templates": [
                {
                    "feature_objects": {
                        "match_mapping_type": "object",
                        "match": "*",
                        "mapping": {
                            "type": "object",
                            "properties": {
                                "feature_value": {"type": "binary"},
                                "value_text": {"type": "text"},
                                "vector_value": {
                                    "type": "dense_vector",
                                    "dims": vector_field_length,
                                    "index": True,
                                    "similarity": config.online_store.similarity,
                                },
                            },
                        },
                    }
                }
            ],
            "properties": {
                "entity_key": {"type": "keyword"},
                "timestamp": {"type": "date"},
                "created_ts": {"type": "date"},
            },
        }

        self._get_client(config).indices.create(
            index=table.name,
            mappings=index_mapping,
        )

    def update(
        self,
        config: RepoConfig,
        tables_to_delete: Sequence[FeatureView],
        tables_to_keep: Sequence[FeatureView],
        entities_to_delete: Sequence[Entity],
        entities_to_keep: Sequence[Entity],
        partial: bool,
    ):
        # implement the update method
        for table in tables_to_delete:
            self._get_client(config).delete_by_query(index=table.name)
        for table in tables_to_keep:
            self.create_index(config, table)

    def teardown(
        self,
        config: RepoConfig,
        tables: Sequence[FeatureView],
        entities: Sequence[Entity],
    ):
        project = config.project
        try:
            for table in tables:
                self._get_client(config).indices.delete(index=table.name)
        except Exception as e:
            logging.exception(f"Error deleting index in project {project}: {e}")
            raise

    def retrieve_online_documents(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: List[str],
        embedding: List[float],
        top_k: int,
        *args,
        **kwargs,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[ValueProto],
            Optional[ValueProto],
            Optional[ValueProto],
        ]
    ]:
        result: List[
            Tuple[
                Optional[datetime],
                Optional[EntityKeyProto],
                Optional[ValueProto],
                Optional[ValueProto],
                Optional[ValueProto],
            ]
        ] = []
        vector_field_path = (
            config.online_store.vector_field_path or "embedding.vector_value"
        )
        query = {
            "script_score": {
                "query": {
                    "bool": {"filter": [{"exists": {"field": vector_field_path}}]}
                },
                "script": {
                    "source": f"cosineSimilarity(params.query_vector, '{vector_field_path}') + 1.0",
                    "params": {"query_vector": embedding},
                },
            }
        }
        body = {"size": top_k, "_source": True, "query": query}
        response = self._get_client(config).search(index=table.name, body=body)
        rows = response["hits"]["hits"][0:top_k]
        for row in rows:
            entity_key = row["_source"]["entity_key"]
            source = row["_source"]
            distance = row["_score"]

            timestamp_str = source.get("timestamp")
            timestamp = datetime.strptime(timestamp_str, "%Y-%m-%dT%H:%M:%S.%f")

            for feature_name in requested_features:
                feature_data = source.get(feature_name, {})
                feature_value = feature_data.get("feature_value")
                vector_value = feature_data.get("vector_value")
                result.append(
                    _build_retrieve_online_document_record(
                        base64.b64decode(entity_key),
                        base64.b64decode(feature_value),
                        str(vector_value),
                        distance,
                        timestamp,
                        config.entity_key_serialization_version,
                    )
                )
        return result

    def retrieve_online_documents_v2(
        self,
        config: RepoConfig,
        table: FeatureView,
        requested_features: List[str],
        embedding: Optional[List[float]],
        top_k: int,
        distance_metric: Optional[str] = None,
        query_string: Optional[str] = None,
    ) -> List[
        Tuple[
            Optional[datetime],
            Optional[EntityKeyProto],
            Optional[Dict[str, ValueProto]],
        ]
    ]:
        """
        Retrieve documents using vector similarity or keyword search from Elasticsearch.
        """
        result: List[
            Tuple[
                Optional[datetime],
                Optional[EntityKeyProto],
                Optional[Dict[str, ValueProto]],
            ]
        ] = []
        if not config.online_store.vector_enabled:
            raise ValueError("Vector search is not enabled in the online store config")

        if embedding is None and query_string is None:
            raise ValueError("Either embedding or query_string must be provided")

        es_index = table.name
        body: Dict[str, Any] = {
            "size": top_k,
        }
        composite_key_name = _get_composite_key_name(table)

        source_fields = requested_features.copy()
        source_fields += ["entity_key", "timestamp"]
        source_fields += composite_key_name
        body["_source"] = source_fields

        if embedding:
            similarity = (distance_metric or config.online_store.similarity).lower()
            vector_field_path = (
                config.online_store.vector_field_path or "embedding.vector_value"
            )
            if similarity == "cosine":
                script = f"cosineSimilarity(params.query_vector, '{vector_field_path}') + 1.0"
            elif similarity == "dot_product":
                script = f"dotProduct(params.query_vector, '{vector_field_path}')"
            elif similarity in ("l2", "l2_norm", "euclidean"):
                script = f"1 / (1 + l2norm(params.query_vector, '{vector_field_path}'))"
            else:
                raise ValueError(
                    f"Unsupported similarity/distance_metric: {similarity}"
                )

        # Hybrid search
        if embedding and query_string:
            body["query"] = {
                "script_score": {
                    "query": {
                        "bool": {
                            "must": [
                                {"query_string": {"query": f'"{query_string}"'}},
                                {"exists": {"field": vector_field_path}},
                            ]
                        }
                    },
                    "script": {
                        "source": script,
                        "params": {"query_vector": embedding},
                    },
                }
            }
        # Vector search only
        elif embedding:
            body["query"] = {
                "script_score": {
                    "query": {
                        "bool": {"filter": [{"exists": {"field": vector_field_path}}]}
                    },
                    "script": {"source": script, "params": {"query_vector": embedding}},
                }
            }
        # Keyword search only
        elif query_string:
            body["query"] = {"query_string": {"query": f'"{query_string}"'}}

        response = self._get_client(config).search(index=es_index, body=body)

        rows = response["hits"]["hits"][0:top_k]
        for row in rows:
            entity_key = row["_source"]["entity_key"]
            entity_key_proto = deserialize_entity_key(
                base64.b64decode(entity_key),
                entity_key_serialization_version=config.entity_key_serialization_version,
            )
            timestamp = row["_source"]["timestamp"]
            timestamp = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")

            # Create feature dict with all requested features
            feature_dict = {"distance": _to_value_proto(float(row["_score"]))}
            if query_string is not None:
                feature_dict["text_rank"] = _to_value_proto(float(row["_score"]))
            join_key_values = _extract_join_keys(entity_key_proto)
            feature_dict.update(join_key_values)

            for feature in requested_features:
                if feature in ("distance", "text_rank"):
                    continue
                value = row["_source"].get(feature, None)
                if value is not None:
                    feature_dict[feature] = _to_value_proto(value)

            result.append((timestamp, entity_key_proto, feature_dict))
        return result


def _to_value_proto(value: Any) -> ValueProto:
    """
    Convert a value to a ValueProto object.
    """
    val_proto = ValueProto()
    if isinstance(value, ValueProto):
        return value
    if isinstance(value, float):
        val_proto.float_val = value
    elif isinstance(value, str):
        val_proto.string_val = value
    elif isinstance(value, int):
        val_proto.int64_val = value
    elif isinstance(value, bool):
        val_proto.bool_val = value
    elif isinstance(value, list) and all(isinstance(v, float) for v in value):
        val_proto.float_list_val.val.extend(value)
    elif isinstance(value, dict) and "feature_value" in value:
        try:
            raw_bytes = base64.b64decode(value["feature_value"])
            val_proto.ParseFromString(raw_bytes)
        except Exception as e:
            raise ValueError(f"Failed to decode feature_value from dict: {e}")
    else:
        raise ValueError(f"Unsupported type for ValueProto: {type(value)}")
    return val_proto


def _encode_feature_value(value: ValueProto) -> Dict[str, Any]:
    """
    Encode a ValueProto into a dictionary for Elasticsearch storage.
    """
    encoded_value = base64.b64encode(value.SerializeToString()).decode("utf-8")
    result = {"feature_value": encoded_value}
    vector_val = get_list_val_str(value)

    if vector_val:
        result["vector_value"] = json.loads(vector_val)
    if value.HasField("string_val"):
        result["value_text"] = value.string_val
    elif value.HasField("bytes_val"):
        result["value_text"] = value.bytes_val.decode("utf-8")
    elif value.HasField("int64_val"):
        result["value_text"] = str(value.int64_val)
    elif value.HasField("double_val"):
        result["value_text"] = str(value.double_val)
    return result


def _get_composite_key_name(table: FeatureView) -> List[str]:
    return [field.name for field in table.entity_columns]


def _extract_join_keys(entity_key_proto) -> Dict[str, ValueProto]:
    join_keys = entity_key_proto.join_keys
    entity_values = entity_key_proto.entity_values
    return {
        join_keys[i]: _to_value_proto(entity_values[i])
        for i in range(len(join_keys))
        if i < len(entity_values)
    }



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: feast/core/Policy.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x66\x65\x61st/core/Policy.proto\x12\nfeast.core\"\xc5\x02\n\x06Policy\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07project\x18\x02 \x01(\t\x12\x38\n\x11role_based_policy\x18\x03 \x01(\x0b\x32\x1b.feast.core.RoleBasedPolicyH\x00\x12:\n\x12group_based_policy\x18\x04 \x01(\x0b\x32\x1c.feast.core.GroupBasedPolicyH\x00\x12\x42\n\x16namespace_based_policy\x18\x05 \x01(\x0b\x32 .feast.core.NamespaceBasedPolicyH\x00\x12S\n\x1f\x63ombined_group_namespace_policy\x18\x06 \x01(\x0b\x32(.feast.core.CombinedGroupNamespacePolicyH\x00\x42\r\n\x0bpolicy_type\" \n\x0fRoleBasedPolicy\x12\r\n\x05roles\x18\x01 \x03(\t\"\"\n\x10GroupBasedPolicy\x12\x0e\n\x06groups\x18\x01 \x03(\t\"*\n\x14NamespaceBasedPolicy\x12\x12\n\nnamespaces\x18\x01 \x03(\t\"B\n\x1c\x43ombinedGroupNamespacePolicy\x12\x0e\n\x06groups\x18\x01 \x03(\t\x12\x12\n\nnamespaces\x18\x02 \x03(\tBP\n\x10\x66\x65\x61st.proto.coreB\x0bPolicyProtoZ/github.com/feast-dev/feast/go/protos/feast/coreb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'feast.core.Policy_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020feast.proto.coreB\013PolicyProtoZ/github.com/feast-dev/feast/go/protos/feast/core'
  _globals['_POLICY']._serialized_start=40
  _globals['_POLICY']._serialized_end=365
  _globals['_ROLEBASEDPOLICY']._serialized_start=367
  _globals['_ROLEBASEDPOLICY']._serialized_end=399
  _globals['_GROUPBASEDPOLICY']._serialized_start=401
  _globals['_GROUPBASEDPOLICY']._serialized_end=435
  _globals['_NAMESPACEBASEDPOLICY']._serialized_start=437
  _globals['_NAMESPACEBASEDPOLICY']._serialized_end=479
  _globals['_COMBINEDGROUPNAMESPACEPOLICY']._serialized_start=481
  _globals['_COMBINEDGROUPNAMESPACEPOLICY']._serialized_end=547
# @@protoc_insertion_point(module_scope)
